####################################################################################################
# neuropythy/math/core.py
# This file contains useful math tools for neuropythy largely built around interfacing seamlessly
# with the PyTorch library.
# by Noah C. Benson

import numpy        as np
import scipy        as sp
import scipy.sparse as sps

# Constants ########################################################################################
pi = np.pi
half_pi = pi / 2
quarter_pi = pi / 4
tau = 2 * pi
inf = np.inf
nan = np.nan
radperdeg = pi / 180
degperrad = 180 / pi

# Importing PyTorch ################################################################################
# We want to work with pytorch but not to require it. Thus we use this function to get it.
def pytorch():
    '''
    Yields the pytorch module or raises an ImportError.
    '''
    global pytorch
    try:
        import torch
        def pytorch():
            '''
            Yields the pytorch module or raises an ImportError.
            '''
            return torch
        return torch
    except Exception: pass
    raise ImportError("failed to import torch: PyTorch may not be installed")

# General Utility Functions ########################################################################
# These are functions that are typically available in both numpy and pytorch; the function here
# provides an interface that works for both numpy and pytorch types.
def to_torchdtype(dtype):
    '''
    to_torchdtype(d) yields d if d is a PyTorch dtype, otherwise yields a dtype
      object that is equivalent to d. If d cannot be converted to a dtype object
      then an exception is raised.

    The argument d may be a numpy dtype object, a numeric type, or a string that
    names a PyTorch dtype.
    '''
    import pimms
    torch = pytorch()
    if isinstance(dtype, torch.dtype):
        return dtype
    elif pimms.is_str(dtype):
        return getattr(torch, dtype)
    elif np.issubdtype(dtype, np.generic):
        return getattr(torch, dtype.__name__)
    else:
        raise ValueError("Cannot convert to pytorch dtype: %s" % (dtype,))
def torchdtype_to_numpydtype(dtype):
    '''
    torchdtype_to_numpydtype(dtype) yields a numpy dtype equivalent to the given
      torch dtype. If dtype is None or if no matching dtype is known, None is
      returned.
    '''
    torch = pytorch()
    dtype = to_torchdtype(dtype)
    return (np.float32    if dtype == torch.float32    else 
            np.float64    if dtype == torch.float64    else
            np.int32      if dtype == torch.int32      else
            np.int64      if dtype == torch.int64      else
            np.float16    if dtype == torch.float16    else
            np.int16      if dtype == torch.int16      else
            np.int8       if dtype == torch.int8       else
            np.bool       if dtype == torch.bool       else
            np.uint8      if dtype == torch.uint8      else
            np.complex64  if dtype == torch.complex64  else
            np.complex128 if dtype == torch.complex128 else
           #np.uint64     if dtype == torch.uint64     else
           #np.uint32     if dtype == torch.uint32     else
           #np.uint16     if dtype == torch.uint16     else
            None)
def isarray(u):
    '''
    isarray(u) yields True if u is either a NumPy array or is a SciPy sparse
      matrix and yields False otherwise.

    See also: istensor(), issparse(), isdense()
    '''
    if   sps.issparse(u):           return True
    elif isinstance(u, np.ndarray): return True
    else: return False
def istensor(u):
    '''
    istensor(u) yields True if u is a PyTorch tensor and False otherwise.
    '''
    try:
        torch = pytorch()
        return torch.is_tensor(u)
    except ImportError: pass
    return False
def issparse(u):
    '''
    issparse(u) yields True if u is either a SciPy sparse matrix or if u is a
      PyTorch sparse tensor; otherwise yields False.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            return u.is_sparse
    except ImportError: pass
    return sps.issparse(u)
def isdense(u):
    '''
    isdense(u) yields True if u is either a dense NumPy array or a dense PyTorch
      tensor and False otherwise.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            return not u.is_sparse
    except ImportError: pass
    if sps.issparse(u): return False
    try: u = np.asarray(u)
    except Exception: return False
    return True
def indices_to_integers(shape, ii):
    '''
    indices_to_integers(shape, ii) yields an integer array whose values
      represent the columns of the index matrix ii (e.g., as returned by
      numpy.where or torch.where).

    The return value is always a numpy array.

    See also integers_to_indices().
    '''
    tr = np.roll(np.cumprod(shape), 1)
    tr[0] = 1
    return np.dot(tr, asarray(ii))
def integers_to_indices(shape, ii):
    '''
    integers_to_indices(shape, ii) yields an indices matrix whose columns
      represent the values of the integer vector ii (e.g., as returned by
      indices_to_integers).

    The return value is always a numpy array.

    See also integers_to_indices().
    '''
    tr = np.roll(np.cumprod(shape), 1)
    tr[0] = 1
    rr = np.zeros((len(shape), len(ii)), dtype=np.int)
    for (rrrow,shval) in zip(reversed(rr), reversed(tr)):
        rrrow[:] = np.floor_divide(ii, shval)
        ii = np.mod(ii, shval)
    return rr
def find_indices(ii, k):
    '''
    find_index(ii, k) yields a NumPy array of the indices at which the
      (integer) values of k occur in the (integer) array ii.
    '''
    n = len(ii)
    zs = np.zeros(n, dtype=np.int)
    ri = sps.csr_matrix((np.arange(n), [k, zs]))
    return np.asarray(ri[ii, zs]).flatten()
def dense_cmp_dense(f, a, b, nannan=False, nanval=False):
    '''
    dense_cmp_dense(f, a, b) yields a dense boolean numpy array of values in the
      elementwise comparison array/tensor represented by the arguments a and b.
      The value of the comparison must be f(a,b) where f is a numpy function
      such as np.equal or np.less_equal.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(a): a = a.detach().numpy()
        if torch.is_tensor(b): b = b.detach().numpy()
    except ImportError: pass
    a = np.asarray(a)
    b = np.asarray(b)
    # This is pretty straightforward!
    r = np.asarray(f(a, b), dtype=np.bool)
    # Check for NaNs
    if nannan or nanval:
        iia = np.isnan(a)
        iib = np.isnan(b)
        if nannan:
            nn = (iia & iib)
            r[nn] = nannan
        if nanval:
            nv = (iia | iib)
            r[nv] = nanval
    # That's it!
    return r
def sparse_cmp_dense(f, ij, dat, x, nannan=False, nanval=False):
    '''
    sparse_cmp_dense(f, ij, dat, x_dense) yields a dense numpy array of the
      elementwise comparison array/tensor representd by the indices and data
      values given and the dense array/tensor x_dense. The ij and dat pairs are
      the reesult of calling either the scipy.sparse.find() function (which
      returns (ij[0], ij[1], dat)) or the indices() (ij) and values() (dat)
      methods on a PyTorch sparse tensor. The value of the comparison must be
      f(a,b) where f is a numpy function such as np.equal or np.less_equal.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    try:
        torch = pytorch()
        if torch.is_tensor( ij1):  ij1 =  ij1.detach().numpy()
        if torch.is_tensor(dat1): dat1 = dat1.detach().numpy()
        if torch.is_tensor(   x):    x =    x.detach().numpy()
    except ImportError: pass
    ij1 = tuple(np.asarray(ij1))
    dat1 = np.asarray(dat1)
    x = np.asarray(x)
    # We start by making a matrix that compares the sparse value to x.
    r = dense_cmp_dense(f, 0, x, nannan=nannan, nanval=nanval)
    # Next, we grab the specific elements with non-zero comparisons to be made.
    xij = x[ij]
    # Compare these...
    rij = dense_cmp_dense(f, dat1, xij, nannan=nannan, nanval=nanval)
    # Add them into the result array.
    r[ij] = rij
    # That's all that is needed.
    return r
def dense_cmp_sparse(f, x, ij, dat, nannan=False, nanval=False):
    '''
    dense_cmp_sparse(f, x_dense, ij, dat) yields a dense numpy array of the
      elementwise comparison array/tensor representd by the indices and data
      values given and the dense array/tensor x_dense. The ij and dat pairs are
      the reesult of calling either the scipy.sparse.find() function (which
      returns (ij[0], ij[1], dat)) or the indices() (ij) and values() (dat)
      methods on a PyTorch sparse tensor. The value of the comparison must be
      f(a,b) where f is a numpy function such as np.equal or np.less_equal.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    try:
        torch = pytorch()
        if torch.is_tensor( ij1):  ij1 =  ij1.detach().numpy()
        if torch.is_tensor(dat1): dat1 = dat1.detach().numpy()
        if torch.is_tensor(   x):    x =    x.detach().numpy()
    except ImportError: pass
    ij1 = tuple(np.asarray(ij1))
    dat1 = np.asarray(dat1)
    x = np.asarray(x)
    # We start by making a matrix that compares the sparse value to x.
    r = dense_cmp_dense(f, x, 0, nannan=nannan, nanval=nanval)
    # Next, we grab the specific elements with non-zero comparisons to be made.
    xij = x[ij]
    # Compare these... (the dense_cmp_dense function handles the NaNs).
    rij = dense_cmp_dense(f, xij, dat1, nannan=nannan, nanval=nanval)
    # Add them into the result array.
    r[ij] = rij
    # That's all that is needed.
    return r
def sparse_cmp_sparse(f, ij1, dat1, ij2, dat2, shape,
                      nannan=False, nanval=False, backend=sps):
    '''
    sparse_cmp_sparse(cmp ij1, dat1, ij2, dat2) yields an array of comparisons
      in the elementwise equality array/tensor representd by the indices and
      data values given. The ij and dat pairs are the reesult of calling either
      the scipy.sparse.find() function (which returns (ij[0], ij[1], dat)) or
      the indices() (ij) and values() (dat) methods on a PyTorch sparse tensor.
      The value of the comparison must be f(a,b) where f is a numpy function
      such as np.equal or np.less_equal.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
      * backend (default: None) the backend library to use for the resulting
        comparison array; should be either torch (PyTorch) or scipy.sparse.
    '''
    # There's no reason not to use numpy arrays; we want to detach torch arrays
    # anyway.
    try:
        torch = pytorch()
        if torch.is_tensor( ij1):  ij1 =  ij1.detach().numpy()
        if torch.is_tensor(dat1): dat1 = dat1.detach().numpy()
        if torch.is_tensor( ij2):  ij2 =  ij2.detach().numpy()
        if torch.is_tensor(dat2): dat2 = dat2.detach().numpy()
    except ImportError: pass
    ij1  = np.asarray(ij1)
    dat1 = np.asarray(dat1)
    ij2  = np.asarray(ij2)
    dat2 = np.asarray(dat2)
    if ij1.shape[0] != ij2.shape[0]:
        raise ValueError('sparse arrays must have equal dimensionality to be compared')
    # Convert the indices into integers by linearizing the cartesian indices.
    tr = np.roll(np.cumprod(shape), 1)
    tr[0] = 1
    k1 = np.dot(tr, ij1)
    k2 = np.dot(tr, ij2)
    # We also need two sparse matrices that reverse these indices.
    ri1 = sps.csr_matrix((np.arange(len(k1)), [k1, np.zeros(len(k1), dtype=np.int)]))
    ri2 = sps.csr_matrix((np.arange(len(k2)), [k2, np.zeros(len(k2), dtype=np.int)]))
    # Figure out which keys get compared with each other versus with 0.
    k_cmp12 = np.intersect1d(k1, k2)
    k_cmp10 = np.setdiff1d(k1, k_cmp12)
    k_cmp02 = np.setdiff1d(k2, k_cmp12)
    # Make the comparisons; regarding NaNs: we pass these values down to the dense_cmp_dense
    # function, so it handles the NaNs for us.
    if len(k_cmp12) > 0:
        zs = np.zeros(len(k_cmp12), dtype=np.int)
        ii1 = np.asarray(ri1[k_cmp12, zs]).flatten()
        ii2 = np.asarray(ri2[k_cmp12, zs]).flatten()
        r_cmp12 = dense_cmp_dense(f, dat1[ii1], dat2[ii2], nannan=nannan, nanval=nanval)
    else:
        r_cmp12 = []
    if len(k_cmp10) > 0:
        ii1 = np.asarray(ri1[k_cmp10, np.zeros(len(k_cmp10), dtype=np.int)]).flatten()
        r_cmp10 = dense_cmp_dense(f, dat1[ii1], 0, nannan=nannan, nanval=nanval)
    else:
        r_cmp10 = []
    if len(k_cmp02) > 0:
        ii2 = np.asarray(ri2[k_cmp02, np.zeros(len(k_cmp02), dtype=np.int)]).flatten()
        r_cmp02 = dense_cmp_dense(f, 0, dat2[ii2], nannan=nannan, nanval=nanval)
    else:
        r_cmp02 = []
    # Create the output values.
    k = np.concatenate([k_cmp12, k_cmp10, k_cmp02])
    ij = np.zeros((len(shape), len(k)), dtype=np.int)
    for (ijrow,shval) in zip(reversed(ij), reversed(tr)):
        ijrow[:] = np.floor_divide(k, shval)
        k = np.mod(k, shval)
    v = np.concatenate([rr for rr in [r_cmp12, r_cmp10, r_cmp02] if len(rr) > 0])
    # We will return a dense or sparse matrix depending on how the sparse values compare.
    if bool(f(0, 0)):
        # We need to return a bunch of ones with the tallied indices set.
        if backend is sps:
            r = np.ones(shape, dtype=np.bool)
            r[tuple(ij)] = v
        else:
            r = torch.ones(shape, dtype=torch.bool)
            r[tuple(ij)] = torch.tensor(v, dtype=torch.bool)
        return r
    else:
        # We return a sparse array of the appropriate type; we can ignore False values (recall that
        # v is a boolean mask already).
        if backend is sps:
            return sps.csr_matrix((v[v], ij[:,v]), shape=shape, dtype=np.bool)
        else:
            return backend.sparse_coo_tensor(ik[:,v], v[v], shape, dtype=backend.bool)
def cmp(f, a, b, nannan=False, nanval=False):
    '''
    cmp(f, a, b) yields f(a, b) as an elementwise comparison of a and b. The
      function f must be a predicate function that works on NumPy arrays, such
      as np.equal or np.less_equal.

    The cmp function works on all NumPy arrays, NumPy-compatible objects (like
    lists and numbers), PyTorch tensors, and SciPy sparse matrices. Note that
    in the case of PyTorch tensors, the result of a cmp call will always be
    detached from the inputs.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    (ash,bsh) = (np.shape(a), np.shape(b))
    shape = ash if len(ash) > len(bsh) else bsh
    try:
        torch = pytorch()
        r = None
        if torch.is_tensor(a):
            if a.is_sparse:
                if not a.is_coalesced(): a = a.coalesce()
                if torch.is_tensor(b):
                    if b.is_sparse:
                        if not b.is_coalesced(): b = b.coalesce()
                        return sparse_cmp_sparse(f, a.indices(), a.values(),
                                                 b.indices(), b.values(), shape,
                                                 nannan=nannan, nanval=nanval,
                                                 backend=torch)
                    else:
                        r = sparse_cmp_dense(f, a.indices(), a.values(), b,
                                             nannan=nannan, nanval=nanval)
                elif sps.issparse(b):
                    (iib,jjb,datb) = sps.find(b)
                    return sparse_cmp_sparse(f, a.indices(), a.values(),
                                             (iib,jjb), datb, shape,
                                             nannan=nannan, nanval=nanval,
                                             backend=torch)
                else:
                    r = sparse_cmp_dense(f, a.indices(), a.values(), b,
                                         nannan=nannan, nanval=nanval)
            elif b.is_sparse:
                if not b.is_coalesced(): b = b.coalesce()
                r = dense_cmp_sparse(f, a, b.indices(), b.values(),
                                     nannan=nannan, nanval=nanval)
            else:
                r = dense_cmp_dense(f, a, b, nannan=nannan, nanval=nanval)
        elif torch.is_tensor(b):
            if b.is_sparse:
                if not b.is_coalesced(): b = b.coalesce()
                if sps.issparse(a):
                    (iia,jja,data) = sps.find(a)
                    return sparse_cmp_sparse(f, (iia,jja), data,
                                             b.indices(), b.values(), shape,
                                             nannan=nannan, nanval=nanval,
                                             backend=torch)
                else:
                    r = dense_cmp_sparse(f, a, b.indices(), b.values(),
                                         nannan=nannan, nanval=nanval)
            elif sps.issparse(a):
                (iia,jja,data) = sps.find(a)
                r = sparse_cmp_dense(f, (iia,jja), data, b,
                                     nannan=nannan, nanval=nanval)
            else:
                r = dense_cmp_dense(f, a, b, nannan=nannan, nanval=nanval)
        if r is not None: return torch.tensor(r)
    except ImportError: pass
    if sps.issparse(a):
        (iia,jja,data) = sps.find(a)
        if sps.issparse(b):
            (iib,jjb,datb) = sps.find(b)
            return sparse_cmp_sparse(f, (iia,jja), data, (iib,jjb), datb, shape,
                                     nannan=nannan, nanval=nanval,
                                     backend=sps)
        else:
            return sparse_cmp_dense(f, (iia,jja), data, b,
                                     nannan=nannan, nanval=nanval)
    elif sps.issparse(b):
        (iib,jjb,datb) = sps.find(b)
        return dense_cmp_sparse(f, a, (iib,jjb), datb,
                                nannan=nannan, nanval=nanval)
    else:
        return dense_cmp_dense(f, a, b, nannan=nannan, nanval=nanval)        
def eq(a, b, nannan=False, nanval=False):
    '''
    eq(a, b) yields (a == b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.equal, a, b, nannan=nannan, nanval=nanval)
def ne(a, b, nannan=True, nanval=True):
    '''
    ne(a, b) yields (a != b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: True) the value returned when comparing NaN against a
        valid value.
      * nannan (default: True) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.not_equal, a, b, nannan=nannan, nanval=nanval)
def lt(a, b, nannan=False, nanval=False):
    '''
    lt(a, b) yields (a < b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.less, a, b, nannan=nannan, nanval=nanval)
def gt(a, b, nannan=False, nanval=False):
    '''
    gt(a, b) yields (a > b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.greater, a, b, nannan=nannan, nanval=nanval)
def le(a, b, nannan=False, nanval=False):
    '''
    le(a, b) yields (a <= b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.less_equal, a, b, nannan=nannan, nanval=nanval)
def ge(a, b, nannan=False, nanval=False):
    '''
    ge(a, b) yields (a >= b) as an elementwise comparison of a and b.

    The arguments a and b can be NumPy-compatible arrays/numbers/lists, SciPy
    sparse matrices, or PyTorch tensors.

    The following options may be given:
      * nanval (default: False) the value returned when comparing NaN against a
        valid value.
      * nannan (default: False) the value returned when comparing NaN against
        another NaN (i.e., is NaN == NaN?).
    '''
    return cmp(np.greater_equal, a, b, nannan=nannan, nanval=nanval)

# Copying and converting tensors/arrays between types ##############################################
def clone(u, **kw):
    '''
    clone(u) yields a copy of u. If u is a pytorch tensor, then this copy is
      not detached.

    If the device keyword argument is passed, then the result is always a
    PyTorch tensor.
    '''
    hasdev = 'device' in kw
    try:
        torch = pytorch()
        if torch.is_tensor(u) or hasdev:
            return totensor(u, **kw)
    except ImportError: pass
    if hasdev:
        raise ValueError("device given but failed to import PyTorch--make sure it is installed")
    if sps.issparse(u):
        return u.copy(**kw)
    else:
        return np.array(u, **kw)
def totensor(u, **kw):
    '''
    totensor(u) yields a copy of u as a PyTorch tensor. Keyword arguments to the
      tensor function may be passed to totensor(). If u is a tensor, a copy of u
      is always returned. Note that u may be a scipy sparse array, in which case
      it is converted into an uncoalesced sparse COO tensor.

    The tensors returned by totensor() are always detached.

    All optional keywords are passed to the torch.tensor() function; however
    the dtype argument is handled specially: if dtype is a string, then the
    dtype that is passed to tensor is getattr(torch, dtype) instead of dtype
    ifself.
    '''
    torch = pytorch()
    dtype = kw.pop('dtype', None)
    if dtype is None:
        npdtype = None
    else:
        dtype = to_torchdtype(dtype)
        npdtype = torchdtype_to_numpydtype(dtype)
    if torch.is_tensor(u):
        if len(kw) == 0 and (dtype is None or dtype == u.dtype):
            return u.clone().detach()
        if u.is_sparse:
            u = u.coalesce()
            # We just detach the indices and values
            v = torch.sparse_coo_tensor(u.indices().clone().detach(), u.values().clone().detach(),
                                        u.shape, dtype=dtype, **kw)
            return v
        # The easiest way to proceed with a dense pytorch tensor is to just start with the
        # numpy array and proceed as usual.
        u = u.detach().numpy()
    # If we have a scipy sparse tensor, we treat this all differently
    if sps.issparse(u):
        (ii,jj,xx) = sps.find(u)
        if npdtype is not None: xx = np.asarray(xx, dtype=npdtype)
        return torch.sparse_coo_tensor([ii,jj], xx, dtype=dtype, **kw)
    else:
        u = np.array(u, dtype=npdtype)
        return torch.tensor(u, dtype=dtype, **kw)
def toarray(u, **kw):
    '''
    toarray(u) yields a copy of u that is either a NumPy array or a SciPy sparse
      matrix that is equivalent to the given array u. This function always
      makes a copy of the given object.
    '''
    dt = kw.pop('dtype', None)
    if dt is not None: kw['dtype'] = to_torchdtype(dt)
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            if u.is_sparse:
                if not u.is_coalesced(): u = u.coalesce()
                ((ii,jj), xx) = (u.indices().detach().numpy(), u.values())
                xx = np.array(xx.detach().numpy(), dtype=dt)
                return sps.coo_matrix((xx, (ii,jj)), **kw)
            u = u.detach().numpy()
    except ImportError: pass
    if sps.issparse(u):
        if len(kw) == 0: return u.copy()
        (ii,jj,xx) = sps.find(u)
        xx = np.array(xx.detach().numpy(), dtype=dt)
        return sps.coo_matrix((xx, (ii,jj)), **kw)
    else:
        return np.array(u, **kw)        
def astensor(u, **kw):
    '''
    astensor(u) yields u if u is a PyTorch tensor and yields a PyTorch-tensor
      version of u otherwise. This is similar to totensor() except that totensor()
      always makes a copy, while astensor() only makes a copy if it must.

    Any keyword arguments for the torch.tensor() function are respected and can
    result in a copy of the tensor being made if they differ from those of the
    torch.tensor() function.
    '''
    torch = pytorch()
    if torch.is_tensor(u):
        # We return u if it is a tensor that has an appropriate dtype and device
        if len(kw) == 0: return u
        if 'dtype' in kw:
            dt = kw['dtype']
            if dt is not None: kw['dtype'] = to_torchdtype(dt)
        return torch.as_tensor(u, **kw)
    else:
        return totensor(u, **kw)
def asarray(u, **kw):
    '''
    asarray(u) yields u if u is a Numpy array; otherwise yields a version of u
      as a Numpy array. A copy of u will not be made unless such a copy is
      required.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            if u.is_sparse:
                if not u.is_coalesced(): u = u.coalesce()
                (ij,uu) = (u.indices(), u.values())
                uu = uu.detach().numpy()
                if len(kw) > 0: uu = np.asarray(uu.detach().numpy(), **kw)
                return sps.coo_matrix((uu, ij.detach().numpy()), u.shape)
            return u.detach().numpy()
    except ImportError: pass
    if sps.issparse(u):
        (ii,jj,uu) = sps.find(u)
        if len(kw) > 0: uu = np.asarray(uu, **kw)
        return torch.sparse_coo_matrix([ii,jj], uu)
    else:
        return np.asarray(u, **kw)
def asdense(u):
    '''
    asdense(u) yields u if u is a dense (non-sparse) array or tensor, asarray(u)
      if u is not an array, scipy matrix, or tensor, and a dense representation
      of u if it is sparse.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            return u.to_dense() if u.is_sparse else u
    except ImportError: pass
    if sps.issparse(u): return u.toarray()
    else: return np.asarray(u)
def todtype(u, dtype):
    '''
    todtype(u, dtype) yields a a copy of PyTorch tensor or NumPy array u with
      the new dtype. The argument u may also be a SciPy sparse array.

    A detached copy of u is always returned. To yield the array itself in cases
    where this is possible, use asdtype(u, dtype).
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            if u.is_sparse:
                u = u.coalesce()
                # We just detach the indices and values
                vals = todtype(u.values(), dtype)
                return torch.sparse_coo_tensor(u.indices().clone().detach(), vals,
                                               u.shape, dtype=dtype, device=u.device)
            else:
                dtype = to_torchdtype(dtype)
                npdtype = torchdtype_to_numpydtype(dtype)
                return torch.tensor(todtype(u.detach().numpy(), npdtype),
                                    dtype=dtype, device=u.device)
    except ImportError: pass
    if sps.issparse(u):
        (ii,jj,xx) = sps.find(u)
        xx = np.array(xx.detach().numpy(), dtype=dtype)
        return sps.coo_matrix((xx, (ii,jj)), **kw)
    else:
        return np.array(u, dtype=dtype)
def asdtype(u, dtype):
    '''
    asdtype(u, dtype) yields either a copy of the PyTorch tensor or NumPy array
      u with the new dtype, if the dtype is not equivalent to u.dtype, or yields
      u itself if the dtype already matches. The argument u may also be a SciPy
      sparse array.

    A copy of u is made only if necessary, and it is not detached unless
    necessary. To always obtain a detached copy, you can use the todtype()
    function.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            dtype = to_torchdtype(dtype)
            if u.dtype == dtype: return u
            if u.is_sparse:
                u = u.coalesce()
                # We just detach the indices and values
                vals = u.values().type(dtype)
                return torch.sparse_coo_tensor(u.indices().clone().detach(), vals,
                                               u.shape, dtype=dtype, device=u.device)
            else:
                npdtype = torchdtype_to_numpydtype(dtype)
                return torch.tensor(asdtype(u.detach().numpy(), npdtype),
                                    dtype=dtype, device=u.device)
    except ImportError: pass
    dtype = np.dtype(dtype)
    if sps.issparse(u):
        if u.dtype == dtype: return u
        (ii,jj,xx) = sps.find(u)
        xx = np.array(xx.detach().numpy(), dtype=dtype)
        return sps.coo_matrix((xx, (ii,jj)), **kw)
    elif np.dtype(u) == dtype:
        return u
    else:
        return np.asarray(u).astype(dtype)
def promote(*args, **kw):
    '''
    promote(a, b) yields a tuple (A, B) where A and B are both the same
      type of array. If either a or b is a PyTorch tensor, then both A and B
      will be PyTorch tensors. Otherwise, they will both be numpy arrays.
    promote(a, b, c...) yields a tuple of all the arguments promoted to
      the appropriate class.

    The optional argument copy (default: False) may be set to true to indicate
    that copies should be made. In this case, copies are made of all arguments
    even if nothing is promoted.
    '''
    copy = kw.pop('copy', False)
    # First, figure out what the promotion type is:
    try:
        torch = pytorch()
        if any(torch.is_tensor(u) for u in args):
            (tof, asf) = (totensor, astensor)
        else:
            (tof, asf) = (toarray, asarray)
    except ImportError:
        (tof, asf) = (toarray, asarray)
    if copy: return tuple([tof(u, **kw) for u in args])
    else:    return tuple([asf(u, **kw) for u in args])
def reshape(u, shape):
    '''
    reshape(u, new_shape) is equivalent to NumPy or PyTorch's reshape functions
      and works on both NumPy arrays and PyTorch tensors as well as SciPy sparse
      matrices.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u): return torch.reshape(u, shape)
    except ImportError: pass
    if sps.issparse(u):
        return u.reshape(shape)
    else:
        return np.reshape(u, shape)

# Access Functions #################################################################################
def extract(u, ii):
    '''
    extract(u, ii) yields an array or tensor of the elements indexed by the
      integer, vector, or matrix ii from the NumPy array, SciPy matrix, or
      PyTorch tensor u.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            if u.is_sparse:
                u = u.coalesce()
                ki = indices_to_integers(u.shape, ii)
                ku = indices_to_integers(u.shape, u.indices())
                kk = np.intersect1d(ki, ku)
                r = torch.zeros(len(ki), dtype=u.dtype, device=u.device)
                r[find_index(ki, kk)] = u.values()[find_index(ku, kk)]
                return r
            else:
                return u[tuple(ii)]
        if torch.is_tensor(ii): ii = ii.detach().numpy()
    except ImportError: pass
    if sps.issparse(u):
        return np.asarray(u[tuple(ii)]).flatten()
    else:
        return np.asarray(u)[tuple(ii)]

# Mathematical Functions ###########################################################################
# These two functions make it easier to handle both PyTorch and Numpy arrays in either unary or
# binary math functioins.
def unary_handle(torch_fn, numpy_fn, zeroval, u, **kw):
    '''
    Handles the unary dispatch of the function represented by torch_fn and
    numpy_Fn for the given object u and the keyword arguments, which are
    passed along to the associated functions.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(u):
            if u.is_sparse:
                if not u.is_coalesced(): u = u.coalesce()
                ((ii,jj), uu) = (u.indices(), u.values())
                if zeroval == 0:
                    vv = torch_fn(uu, **kw)
                    return torch.sparse_coo_matrix((ii,jj), vv, dtype=u.dtype, device=u.device)
                else:
                    r = torch.full(u.shape, zeroval, dtype=u.dtype, device=u.device)
                    r[ii,jj] = torch_fn(uu, **kw)
                    return r
            else:
                return torch_fn(u, **kw)
    except ImportError: pass
    if sps.issparse(u):
        if zeroval == 0:
            r = u.copy()
            r.data[:] = numpy_fn(r.data, **kw)
        else:
            r = np.full(u.shap, zeroval, dtype=u.dtype)
            (ii,jj,uu) = sps.find(u)
            r[ii,jj] = numpy_fn(uu, **kw)
        return r
    else:
        return numpy_fn(u, **kw)
def binary_handle(torch_fn, numpy_fn, a, b, **kw):
    '''
    Handles binary operators.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse: a = a.to_dense()
            if torch.is_tensor(b):
                if b.is_sparse: b = b.to_dense()
            else:
                b = totensor(b)
            return torch_fn(a, b, **kw)
        elif torch.is_tensor(b):
            a = totensor(a)
            return torch_fn(a, b, **kw)
    except ImportError: pass
    if sps.issparse(a): a = a.toarray()
    if sps.issparse(b): b = b.toarray()
    return numpy_fn(a, b, **kw)
def unary_mathfn(numpy_fn):
    '''
    @unary_mathfn(numpy_fn) is a decorator that allows the PyTorch-specific
    function that follows it to be compatible with PyTorch sprase arrays,
    scipy sparse arrays, numpy arrays, python lists, and numbers.
    '''
    import functools, warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        zeroval = numpy_fn(0.0)
    def wrap1(torch_fn):
        @functools.wraps(torch_fn)
        def wrapper(u, **kw):
            return unary_handle(torch_fn, numpy_fn, zeroval, u, **kw)
        return wrapper
    return wrap1
def binary_mathfn(numpy_fn):
    '''
    @binary_mathfn(numpy_fn) is a decorator that allows the PyTorch-specific
    function that follows it to be compatible with PyTorch sprase arrays,
    scipy sparse arrays, numpy arrays, python lists, and numbers.
    '''
    import functools
    def wrap1(torch_fn):
        @functools.wraps(torch_fn)
        def wrapper(u, **kw):
            return binary_handle(torch_fn, numpy_fn, u, **kw)
        return wrapper
    return wrap1
def summate(a, b):
    '''
    summate(a, b) yields a + b and works correctly for NumPy arrays, PyTorch
      tensors (sparse and dense), SciPy sparse matrices, and and general numbers
      and lists.

    Generally speaking, the add() function is preferred.
    '''
    (a, b) = promote(a, b)
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse:
                if b.is_sparse:
                    return a + b
                else:
                    (b, a) = (a, b)
            if b.is_sparse:
                (ijb, xb) = (tuple(b.indices()), a.values())
                a = a + torch.zeros((), dtype=b.dtype, device=b.device)
                a[ijb] += xb
                return a
            else:
                return a + b
    except ImportError: pass
    if sps.issparse(a) or sps.issparse(b):
        return np.asarray(a + b)
    else:
        return a + b
def add(*args):
    '''
    add(a, b, ...) yields the sum of all the given values. This is equivalent to
      (a + b + ...) except that it works for NumPy arrays, SciPy sparse
      matrices, and PyTorch tensors (including sparse tensors).
    '''
    from functools import reduce
    if   len(args) == 0: return 0
    elif len(args) == 1: return args[0]
    else:                return reduce(summate, *args)
def sub(a, b):
    '''
    sub(a, b) yields the a minus b. This is equivalent to (a - b) except
      that it works for NumPy arrays, SciPy sparse matrices, and PyTorch tensors
      (including sparse tensors).
    '''
    (a, b) = promote(a, b)
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse and not b.is_sparse: a = a.to_dense()
            return a - b
    except ImportError: pass
    if sps.issparse(a) or sps.issparse(b):
        return np.asarray(a - b)
    else:
        return a - b
def multiply(a, b):
    '''
    multiply(a, b) yields a * b and works correctly for NumPy arrays, PyTorch
      tensors (sparse and dense), SciPy sparse matrices, and and general numbers
      and lists.
    '''
    (a, b) = promote(a, b)
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse:
                if b.is_sparse:
                    return a * b
                else:
                    (ija, xa) = (a.indices(), a.values())
                    xb = b[tuple(ija)]
                    xa = xa * xb
                    nans = torch.where(~torch.isfinite(b))
                    if len(nans[0]) > 0:
                        ija = torch.stack([torch.cat(ij) for ij in zip(ija, nans)])
                        nans = torch.full(xa.shape, np.nan, dtype=xa.dtype, device=xa.device)
                        xa = torch.cat((xa, nans))
                    return torch.sparse_coo_tensor(ija, xa, dtype=xa.dtype, device=xa.device)
            elif b.is_sparse:
                (ijb, xb) = (b.indices(), a.values())
                xa = a[tuple(ijb)]
                xb = xa * xb
                nans = torch.where(~torch.isfinite(a))
                if len(nans[0]) > 0:
                    ijb = torch.stack([torch.cat(ij) for ij in zip(ijb, nans)])
                    nans = torch.full(xa.shape, np.nan, dtype=xb.dtype, device=xb.device)
                    xa = torch.cat((xb, nans))
                return torch.sparse_coo_tensor(ijb, xb, dtype=xb.dtype, device=xb.device)
            else:
                return a * b
    except ImportError: pass
    if sps.issparse(a):
        return a.multiply(b)
    elif sps.issparse(b):
        return b.multiply(a)
    else:
        return a * b
def mul(*args):
    '''
    mul(a, b, ...) yields the product of all the given values. This is
      equivalent to (a * b * ...) except that it works for NumPy arrays, SciPy
      sparse matrices, and PyTorch tensors (including sparse tensors).
    '''
    from functools import reduce
    if   len(args) == 0: return 1
    elif len(args) == 1: return args[0]
    else:                return reduce(multiply, *args)
def div(a, b):
    '''
    div(a, b) yields the a divided by b. This is equivalent to (a / b) except
      that it works for NumPy arrays, SciPy sparse matrices, and PyTorch tensors
      (including sparse tensors).
    '''
    (a, b) = promote(a, b)
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse: a = a.to_dense()
            if b.is_sparse: b = b.to_dense()
            return a / b
    except ImportError: pass
    if sps.issparse(a):
        return np.asarray(a / b)
    elif sps.issparse(b):
        return (a / b.toarray())
    else:
        return a / b
def mod(a, b):
    '''
    mod(a, b) yields a mod b. The arguments may be PyTorch tensors, numpy
      arrays, SciPy sparse matrices, or primitives. The PyTorch function used is
      remainder.
    '''
    (a,b) = promote(a, b)
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            if a.is_sparse: a = a.to_dense()
            if b.is_sparse: b = b.to_dense()
            return torch.remainder(a, b)
    except ImportError: pass
    if sps.issparse(a): a = asdense(a)
    if sps.issparse(b): b = asdense(b)
    return np.mod(a, b)
def _numpy_safesqrt(x):
    u = np.array(x)
    ii = u > 0
    u[u < 0] = x
    s = u[ii]
    s[:] = np.sqrt(s)
    return u
@unary_mathfn(_numpy_safesqrt)
def safesqrt(u, x=0):
    '''
    safesqrt(u, x) is equivalent to torch.sqrt(u) but only operates on values
      that are greater than or equal to 0; elsewhere the values returned is x.
    safesqrt(u) is equivalent to safesqrt(u, 0).
    '''
    torch = pytorch()
    r = torch.full(u.shape, x, dtype=u.dtype, device=u.device)
    ii = u > 0
    r[ii] = torch.sqrt(u[ii])
    return r
@unary_mathfn(np.sqrt)
def sqrt(u):
    '''
    sqrt(u) is equivalent to torch.sqrt(u) or np.sqrt(u) depending on the type
      of u.
    '''
    torch = pytorch()
    return torch.sqrt(u)
@unary_mathfn(np.exp)
def exp(x):
    '''
    exp(x) yields the exponential of x, and works for either PyTorch tensors or
      for numpy arrays.
    '''
    torch = pytorch()
    return torch.exp(x)
@unary_mathfn(np.log)
def log(x):
    '''
    log(x) yields the logarithm of x, and works for either PyTorch tensors or
      for numpy arrays.
    '''
    torch = pytorch()
    return torch.log(x)
@unary_mathfn(np.log10)
def log10(x):
    '''
    log10(x) yields the base-10 logarithm of x, and works for either PyTorch
      tensors or for numpy arrays.
    '''
    torch = pytorch()
    return torch.log10(x)
@unary_mathfn(np.log2)
def log2(x):
    '''
    log2(x) yields the base-2 logarithm of x, and works for either PyTorch
      tensors or for numpy arrays.
    '''
    torch = pytorch()
    return torch.log2(x)
@unary_mathfn(np.abs)
def abs(x):
    '''
    abs(x) yields the absolute value of x, and works for either PyTorch tensors
      or for numpy arrays.
    '''
    torch = pytorch()
    return torch.abs(x)
@unary_mathfn(np.arcsin)
def arcsin(x):
    '''
    arcsin(x) is equivalent to torch.asin(x) or numpy.arcsin(x) depending on
      whether x is a pytorch tensor or not.
    '''
    torch = pytorch()
    return torch.asin(x)
@unary_mathfn(np.arccos)
def arccos(x):
    '''
    arccos(x) is equivalent to torch.arccos(x) or numpy.arccos(x) depending on
      whether x is a pytorch tensor or not.
    '''
    torch = pytorch()
    return torch.acos(x)
@unary_mathfn(np.sin)
def sin(x):
    '''
    sin(x) is equivalent to torch.sin(x) or numpy.sin(x) depending on whether x
      is a pytorch tensor or not.
    '''
    torch = pytorch()
    return torch.sin(x)
@unary_mathfn(np.cos)
def cos(x):
    '''
    cos(x) is equivalent to torch.cos(x) or numpy.cos(x) depending on whether x
      is a pytorch tensor or not.
    '''
    torch = pytorch()
    return torch.cos(x)
@unary_mathfn(np.cos)
def tan(x):
    '''
    tan(x) is equivalent to torch.tan(x) or numpy.tan(x) depending on whether x
      is a pytorch tensor or not.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(x): return torch.tan(x)
    except ImportError: pass
    return np.tan(x)
@unary_mathfn(np.cos)
def lgamma(x):
    '''
    lgamma(x) yields the log-gamma of x and is equivalent to torch.lgamma(x) or
      scipy.special.loggamma(x) depending on whether x is a pytorch tensor or
      not.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(x): return torch.lgamma(x)
    except ImportError: pass
    return sp.special.loggamma(x)
@unary_mathfn(np.arctan)
def atan(y):
    '''
    atan(u) is equivalent to np.arctan(u) or torch.atan(u) depending on the type of u.
    '''
    torch = pytorch()
    if y.is_sparse:
        (ij,x) = (y.indices(), y.values())
        return torch.sparse_coo_tensor(ij, torch.atan(x), dtype=y.dtype, device=y.device)
    else:
        return torch.atan(y)
def atan2(y, x):
    '''
    atan2(y, x) is equivalent to np.arctan2(y,x) or torch.atan2(y,x) depending on the
      types of y and x.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(y) or torch.is_tensor(x):
            if not torch.is_tensor(y): y = astensor(y)
            if not torch.is_tensor(x): x = astensor(x)
            if y.is_sparse:
                ((ii,jj),yy) = (y.indices(), y.values())
                xx = x[ii,jj]
                return torch.sparse_coo_tensor((ii,jj), torch.atan2(yy,xx),
                                               dtype=y.dtype, device=y.device)
            elif x.is_sparse:
                return torch.atan2(y, x.to_dense())
            else:
                return torch.atan2(y, x)
    except ImportError: pass
    if sps.issparse(y):
        (ii,jj,yy) = sps.find(y)
        if sps.issparse(x): xx = np.asarray(x[ii,jj]).flatten()
        else:
            x = np.asarray(x)
            nsh = len(x.shape)
            if   nsh == 0: xx = x
            elif nsh == 1: xx = x[jj]
            else:          xx = x[...,ii,jj]
        r = y.copy()
        r[ii,jj] = np.arctan(yy, xx)
        return r
    elif sps.issparse(x):
        y = np.asarray(y)
        nsh = len(y.shape)
        r = np.arctan(y, x.dtype.type(0))
        (ii,jj,xx) = sps.find(x)
        if   nsh == 0: yy = y
        elif nsh == 1: yy = y[jj]
        else:          yy = y[...,ii,jj]
        r[ii,jj] = np.arctan(yy, xx)
        return r
    else:
        return np.arctan2(y, x)
def arctan(y, x=None):
    '''
    arctan(y, x) is equivalent to torch.atan2(y, x) or numpy.arctan2(y, x)
      depending on whether y and x are pytorch tensors or not.
    arctan(y) is equvalent to torch.atan(y) or numpy.arctan(y).
    '''
    if x is None: return atan(y)
    else:         return atan2(y, x)
def pow(a, b):
    '''
    pow(a, b) yields a**b and works whether a and b are PyTorch tensors, NumPy
      arrays (or compatible objects like numbers), or SciPy sparse arrays.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            b = astensor(b)
            if b.is_sparse: b = b.to_dense()
            if a.is_sparse: a = a.to_dense()
            return a ** b
        elif torch.is_tensor(b):
            a = astensor(a)
            return torch.sqrt(a**2 + b**2)
    except ImportError: pass
    if sps.issparse(a): a = a.toarray()
    else: a = np.asarray(a)
    if sps.issparse(b): b = b.toarray()
    else: b = np.asarray(b)
    return a ** b
def hypot(a, b):
    '''
    hypot(a, b) yields sqrt(a**2 + b**2) and works whether a and b are PyTorch
      tensors or numpy arrays.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            b = astensor(b)
            return torch.sqrt(a**2 + b**2)
        elif torch.is_tensor(b):
            a = astensor(a)
            return torch.sqrt(a**2 + b**2)
    except ImportError: pass
    if sps.issparse(a):
        a2 = a.multiply(a)
        if sps.issparse(b):
            b2 = b.multiply(b)
            return np.sqrt(a2 + b2)
        else:
            b = asarray(b)
            return np.sqrt(a2 + b**2)
    elif sps.issparse(b):
        b2 = b.multiply(b)
        a = asarray(a)
        return np.sqrt(a**2 + b2)
    else:
        (a, b) = (np.asarray(a), np.asarray(b))
        return np.sqrt(a**2 + b**2)
def hypot2(a, b):
    '''
    hypot2(a, b) yields a**2 + b**2 and works whether a and b are PyTorch
      tensors or numpy arrays.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            b = astensor(b)
            return (a**2 + b**2)
        elif torch.is_tensor(b):
            a = astensor(a)
            return (a**2 + b**2)
    except ImportError: pass
    if sps.issparse(a):
        a2 = a.multiply(a)
        if sps.issparse(b):
            b2 = b.multiply(b)
            return (a2 + b2)
        else:
            b = asarray(b)
            return (a2 + b**2)
    elif sps.issparse(b):
        b2 = b.multiply(b)
        a = asarray(a)
        return (a**2 + b2)
    else:
        (a, b) = (np.asarray(a), np.asarray(b))
        return (a**2 + b**2)
def triarea(a,b,c):
    '''
    triarea(a,b,c) yields the area of the triangle whose sides have lengths
      a, b, and c. The arguments may be PyTorch tensors, numpy arrays, or
      Python primitives.
    '''
    hp = 0.5 * add(a, b, c)
    return mul(hp, sub(hp, a), sub(hp, b), sub(hp, c))
def eudist2(a, b):
    '''
    eudist2(a, b) yields the square distance between points a and b. The points
      may be vectors of the same length or matrices with one row per dimension.
      PyTorch tensors, numpy arrays, and primitives are allowed.
    '''
    (a, b) = promote(a, b)
    a = asdense(a)
    b = asdense(b)
    tot = (b[0] - a[0])**2
    for ii in range(1, a.shape[0]):
        tot += (b[ii] - a[ii])**2
    return tot
def eudist(a, b):
    '''
    eudist(a, b) yields the distance between points a and b. The points may be
      vectors of the same length or matrices with one row per dimension. PyTorch
      tensors, numpy arrays, and primitives are allowed.
    '''
    return sqrt(eudist2(a, b))
def trisides2(a, b, c):
    '''
    trisides2(a, b, c) yields the squared lengths of the sides of the triangles
      formed by the given coordinates. The arguemnts a, b, and c may be vectors
      or matrices with rows representing dimensions. PyTorch tensors, numpy
      arrays, and primitives are allowed.
    '''
    return (dist2(b, c), dist2(c, a), dist2(a, b))
def trisides(a, b, c):
    '''
    trisides(a, b, c) yields the lengths of the sides of the triangles formed by
      the given coordinates. The arguemnts a, b, and c may be vectors or
      matrices with rows representing dimensions. PyTorch tensors, numpy arrays,
      and primitives are allowed.
    '''
    (a2,b2,c2) = trisides2(a, b, c)
    return (sqrt(a2), sqrt(b2), sqrt(c2))
def trialtitudes(sides, area=None):
    '''
    trialtitudes((a,b,c)) yields (ha, hb, hc), the altitudes of each
      side associated with the side-lengths, a,b,c.
    trialtitudes((a,b,c), area) uses the given triangle area (which then doesn't
      have to be calculated).
    '''
    (a,b,c) = asdense(sides)
    if area is None: area = triarea(a, b, c)
    return (2*area/a, 2*area/b, 2*area/c)
def rangemod(u, min, max):
    '''
    rangemod(u, min, max) yields mod(u + min, max - min) - min.
    '''
    (u, min, max) = [asdense(x) for x in promote(u, min, max)]
    rng = max - min
    return mod(u + min, rng) - min
def radmod(u):
    '''
    radmod(u) yields mod(u + pi, 2 pi) - pi.
    '''
    return rangemod(u, -pi, pi)
def degmod(u):
    '''
    degmod(u) yields mod(u + 180, 360) - 180.
    '''
    return rangemod(u, -180, 180)
def branch(iftensor, thentensor, elsetensor=None):
    '''
    branch(q, t, e) yields, elementwise for the given tensors, t if q else e.
    branch(q, t) or branch(q, t, None) yields, elementwise, t if q else 0.
    branch(q, None, e) yields, elementwise, 0 if q else e.
    
    The output tensor will always have the same shape as q. The values for t
    and e may be constants or tensors the same shape as q.

    This function should be safe to use in optimization, i.e., for gradient
    calculatioins.
    '''
    if thentensor is None: thentensor = 0.0
    if elsetensor is None: elsetensor = 0.0
    (q, t, e) = promote(iftensor, thentensor, elsetensor)
    q = asdense(q)
    res = clone(q, dtype=t.dtype)
    q = (q != 0) # to convert to boolean
    res[:] = e
    if t.shape == ():        res[q] = t
    elif t.shape == q.shape: res[q] = t[q]
    else:
        ii = np.where(q)
        res[ii] = t[ii[-len(t.shape):]]
    return res
def zinv(x):
    '''
    zinv(x) yields 0 if x == 0 and 1/x otherwise. This is done in a way that
      is safe for torch gradients; i.e., the gradient for any element of x that
      is equal to 0 will also be 0. PyTorch tensors and numpy arrays are
      allowed.
    '''
    try:
        torch = pytorch()
        if torch.is_tensor(x):
            if x.is_sparse:
                (ij, vals) = (x.indices(), x.values())
                vals = zinv(vals)
                return torch.sparse_coo_tensor(ij, vals, dtype=x.dtype, device=x.device)
            ii = (x != 0)
            r = torch.zeros(x.shape, dtype=x.dtype, device=x.device)
            r[ii] = 1 / x[ii]
            return r
    except ImportError: pass
    if sps.issparse(x):
        r = x.copy()
        r.data[:] = zinv(r.data)
        return r
    else:
        x = np.asarray(x)
        ii = (x != 0)
        r = np.zeros(x.shape, dtype=x.dtype)
        r[ii] = 1 / x[ii]
        return r

def numel(x):
    '''
    numel(x) yields the number of elements in x: the product of the shape of x.
    '''
    return int(np.prod(np.shape(x)))
def rows(x):
    '''
    rows(x) yields the number of rows in x; if x is a scalar, this is still 1.
    '''
    s = np.shape(x)
    return s[0] if len(s) > 0 else 1
def check_sparsity(x, fraction=0.6):
    '''
    check_sparsity(x) yields either x or an array equivalent to x with a different sparsity based on
      a heuristic: if x is a sparse array with more than 60% of its elements specified, it is made
      dense; otherwise, it is left alone.

    The optional argument fraction (default 0.6) specifies the fraction of elements that must be
    specified in the array for it to be un-sparsified.
    '''
    if sps.issparse(x):
        n = numel(x)
        if n == 0: return x
        elif len(x.data) / float(n) > 0.6: return x.toarray()
        else: return x
    else:
        try:
            torch = pytorch()
            if torch.is_tensor(x) and x.is_sparse:
                n = numel(x)
                if len(x.indices()) / float(n) > 0.6: return x.to_dense()
        except ImportError: pass
        return x
def unbroadcast(a, b):
    '''
    unbroadcast(a, b) yields a tuple (aa, bb) that is equivalent to (a, b) except that aa and bb
      have been reshaped such that arithmetic numpy operations such as aa * bb will result in
      row-wise operation instead of column-wise broadcasting.
    '''
    # First: promote the arrays.
    (a,b) = [asdense(x) for x in promote(a, b)]
    # See if they're tensors first
    try:
        torch = pytorch()
        if torch.is_tensor(a):
            da = len(a.shape)
            db = len(b.shape)
            if   da > db: return (a, torch.reshape(b, b.shape + (1,)*(da - db)))
            elif da < db: return (torch.reshape(a, a.shape + (1,)*(db - da)), b)
            else:         return (a, b)
    except ImportError: pass
    # they could be sparse:
    spa = sps.issparse(a)
    spb = sps.issparse(b)
    if   spa and spb: return (a,b)
    elif spa or  spb:
        def fix(sp,nm):
            nm = np.asarray(nm)
            dnm = len(nm.shape)
            nnm = np.prod(nm.shape)
            # if we have (sparse matrix) * (high-dim array), unbroadcast the dense array
            if   dnm == 0: return (sp, np.reshape(nm, (1,   1)))
            elif dnm == 1: return (sp, np.reshape(nm, (nnm, 1)))
            elif dnm == 2: return (sp, nm)
            else:          return unbroadcast(sp.toarray(), nm)
        return fix(a, b) if spa else tuple(reversed(fix(b, a)))
    # okay, no sparse matrices found:
    da = len(a.shape)
    db = len(b.shape)
    if   da > db: return (a, np.reshape(b, b.shape + (1,)*(da - db)))
    elif da < db: return (np.reshape(a, a.shape + (1,)*(db - da)), b)
    else:         return (a, b)

# Accumulators #####################################################################################
@unary_mathfn(np.sum)
def sum(x, *args, **kw):
    '''
    sum(x) yields the sum of the elements in x.
    sum(x, axis) yields the sum along the given axis.
    '''
    return pytorch().sum(x, *args, **kw)
@unary_mathfn(np.prod)
def prod(x, *args, **kw):
    '''
    prod(x) yields the product of the elements in x.
    prod(x, axis) yields the product along the given axis.
    '''
    return pytorch().prod(x, *args, **kw)
@unary_mathfn(np.mean)
def mean(x, *args, **kw):
    '''
    mean(x) yields the mean of the elements in x.
    mean(x, axis) yields the mean along the given axis.
    '''
    return pytorch().mean(x, *args, **kw)
@unary_mathfn(np.var)
def var(x, *args, **kw):
    '''
    var(x) yields the variance of the elements in x.
    var(x, axis) yields the variance along the given axis.
    '''
    return pytorch().var(x, *args, **kw)
@unary_mathfn(np.std)
def std(x, *args, **kw):
    '''
    std(x) yields the standard deviation of the elements in x.
    std(x, axis) yields the standard deviation along the given axis.
    '''
    return pytorch().std(x, *args, **kw)
@unary_mathfn(np.median)
def median(x, *args, **kw):
    '''
    median(x) yields the median of the elements in x.
    median(x, axis) yields the median along the given axis.
    '''
    return pytorch().median(x, *args, **kw)
@unary_mathfn(np.min)
def min(x, *args, **kw):
    '''
    min(x) yields the minimum of the elements in x.
    min(x, axis) yields the minimum along the given axis.
    '''
    return pytorch().min(x, *args, **kw)[0]
@unary_mathfn(np.max)
def max(x, *args, **kw):
    '''
    max(x) yields the maximum of the elements in x.
    max(x, axis) yields the maximum along the given axis.
    '''
    return pytorch().max(x, *args, **kw)[0]
@unary_mathfn(np.argmin)
def argmin(x, *args, **kw):
    '''
    argmin(x) yields the inidex of the minimum of the element in x.
    argmin(x, axis) yields the indices of the minima along the given axis.
    '''
    return pytorch().min(x, *args, **kw)[1]
@unary_mathfn(np.argmax)
def argmax(x, *args, **kw):
    '''
    argmax(x) yields the inidex of the maximum of the element in x.
    argmax(x, axis) yields the indices of the maxima along the given axis.
    '''
    return pytorch().max(x, *args, **kw)[1]
@unary_mathfn(np.all)
def all(x, *args, **kw):
    '''
    min(x) yields the minimum of the elements in x.
    min(x, axis) yields the minimum along the given axis.
    '''
    return pytorch().all(x, *args, **kw)


# Distributions ####################################################################################
def beta_log_prob(x, mu, scale):
    '''
    beta_log_prob(mu, scale, x) yields the log probability density of the beta
      distribution parameterized using the mean mu (which must be between 0 and
      1) and scale (which may be any real number) at the value x.
      
    The traditional beta distribution uses parameters a and b. The
    reparameterizatoin of mu and scale here is as follows:
      a = (mu * (2 - b) - 1) / (mu - 1)
      b = (2 - 1/mu) + exp(scale)
    '''
    (x, mu, scale) = promote(x, mu, scale)
    try:
        torch = pytorch()
        if torch.is_tensor(x):
            from torch.distributions import Beta
            b = torch.exp(scale)
            a = (mu * (b - 2) + 1) / (1 - mu)
            dist = Beta(a, b)
            return dist.log_prob(x)
    except ImportError: pass
    from scipy.stats import beta
    b = np.exp(scale)
    a = (mu * (b - 2) + 1) / (1 - mu)
    return beta.logpdf(x, a, b)    
def beta_prob(x, mu, scale):
    '''
    beta_prob(x, mu, scale) yields the probability density function of the beta
      distribution parrameterized using the mean mu (which must be between 0 and
      1) and scale (which may be any real number) at the value x.
    
    The traditional beta distribution uses parameters a and b. The
    reparameterizatoin of mu and scale here is as follows:
      a = (mu * (2 - b) - 1) / (mu - 1)
      b = (2 - 1/mu) + exp(scale)
    '''
    (x, mu, scale) = promote(x, mu, scale)
    try:
        torch = pytorch()
        if torch.is_tensor(x):
            return torch.exp(beta_log_prob(x, mu, scale))
    except ImportError: pass
    from scipy.stats import beta
    b = np.exp(scale)
    a = (mu * (b - 2) + 1) / (1 - mu)
    return beta.pdf(x, a, b)
def normal_log_prob(t, center=0, width=1):
    '''
    normal_log_prob(t) yields the log-probability of the normal 
      distribution: -(t^2 + log(2 pi)) / 2.
    normal_log_prob(t, w) is equivalent to normal_log_prob(t) with a mean of w;
      equal to normal_log_prob(t - w).
    normal_log_prob(t, w, s) uses a standard deviation of s, equivalent to
      normal_log_prob(t - w, s) - log(w)
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            from torch.distributions import Normal
            dist = Normal(center, width)
            return dist.log_prob(t)
    except ImportError: pass
    from scipy.stats import norm
    return norm.logpdf(t, center, width)
def normal_prob(t, center=0, width=1):
    '''
    normal_prob(t) yields the probability density of the normal distribution at
      t: exp(-t^2 / 2) / sqrt(2 pi).
    normal_prob(t, w) is equivalent to normal_log_prob(t) with a mean of w;
      equal to normal_prob(t - w).
    normal_prob(t, w, s) uses a standard deviation of s, equivalent to
      normal_prob(t - w, s) / s.
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            return torch.exp(normal_log_prob(t, center, width))
    except ImportError: pass
    from scipy.stats import norm
    return norm.pdf(t, center, width)
def cauchy_log_prob(t, center=0, width=1):
    '''
    cauchy_log_prob(t) yields the log-probability of the Cauchy
      distribution: -log(pi (1 + t^2)).
    cauchy_log_prob(t, t0) is equivalent to cauchy_log_prob(t-t0).
    cauchy_log_prob(t, t0, s) is equivalent to cauchy_log_prob(t - t0) using
      a distribution with a width of s, equal to: -log(pi w (1 + (t/w)^2)).
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            from torch.distributions import Cauchy
            dist = Cauchy(center, width)
            return dist.log_prob(t)
    except ImportError: pass
    from scipy.stats import cauchy
    return cauchy.logpdf(t, center, width)
def cauchy_prob(t, center=0, width=1):
    '''
    cauchy_prob(t) yields the probability density of the Cauchy distribution at
      t with a scale of 1 and a center of 0.
    cauchy_prob(t, t0) is equivalent to cauchy_prob(t-t0).
    cauchy_prob(t, t0, s) is equivalent to cauchy_prob(t - t0) using
      a distribution with a width of s.
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            return torch.exp(cauchy_log_prob(t, center, width))
    except ImportError: pass
    from scipy.stats import cauchy
    return cauchy.pdf(t, center, width)
def halfcauchy_log_prob(t, width=1):
    '''
    halfcauchy_log_prob(t) yields the log-probability of the half-Cauchy
      distribution.
    halfcauchy_log_prob(t, s) uses a distribution with a width of s, equal to.
    '''
    (t, width) = promote(t, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            from torch.distributions import HalfCauchy
            dist = HalfCauchy(width)
            return dist.log_prob(t)
    except ImportError: pass
    from scipy.stats import halfcauchy
    return halfcauchy.logpdf(t, 0, width)
def halfcauchy_prob(t, width=1):
    '''
    halfcauchy_prob(t) yields the probability density of the half-Cauchy
      distribution at t with a scale of 1.
    halfcauchy_prob(t, s) uses a scale of s.
    '''
    (t, width) = promote(t, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            return torch.exp(halfcauchy_log_prob(t, width))
    except ImportError: pass
    from scipy.stats import halfcauchy
    return halfcauchy.pdf(t, 0, width)
def laplace_log_prob(t, width=1, center=0):
    '''
    laplace_log_prob(t) yields the log-probability of the Laplace
      distribution: -(|t| + log(2))
    laplace_log_prob(t, t0) is equivalent to laplace_log_prob(t - t0).
    laplace_log_prob(t, t0, s) uses a center of t0 and a scale of s.
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            from torch.distributions import Laplace
            dist = Laplace(center, width)
            return dist.log_prob(t)
    except ImportError: pass
    from scipy.stats import laplace
    return laplace.logpdf(t, center, width)
def laplace_prob(t, width=1, center=0):
    '''
    laplace_prob(t) yields the probability density of the Laplace distribution
      at t.
    laplace_prob(t, t0) is equivalent to laplace_prob(t - t0).
    laplace_prob(t, t0, s) uses a center of t0 and a scale of s.
    '''
    (t, center, width) = promote(t, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            return torch.exp(laplace_log_prob(t, center, width))
    except ImportError: pass
    from scipy.stats import laplace
    return laplace.pdf(t, center, width)
def exp_log_prob(t, width=1):
    '''
    exp_log_prob(t) yields the log-probability of the exponential distribution.
    exp_log_prob(t, s) uses a distribution with a width of s.
    '''
    (t, width) = promote(t, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            from torch.distributions import Exponential
            dist = Exponential(1 / width)
            return dist.log_prob(t)
    except ImportError: pass
    from scipy.stats import expon
    return expon.logpdf(t, 0, width)
def exp_prob(t, width=1):
    '''
    exp_prob(t) yields the probability density of the exponential distribution
      at t with a scale of 1.
    exp_prob(t, s) uses a scale of s.
    '''
    (t, width) = promote(t, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            return torch.exp(exp_log_prob(t, width))
    except ImportError: pass
    from scipy.stats import expon
    return expon.pdf(t, 0, width)
def gennorm_log_prob(t, q, center=0, width=1):
    '''
    gennorm_log_prob(t, q) yields the log-probability of the generalized error
      distribution, whose PDF is given by q / (2 gamma(1 / q)) exp(-|t|^q).
    genorm_log_prob(t, q, t0) is equivalent to gennorm_log_prob(t - t0, q).
    genorm_log_prob(t, q, t0, s) uses the scale s.
    '''
    (t, q, center, width) = promote(t, q, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            const = q / (2 * lgamma(1 / q)) / width
            return torch.log(const) - abs((t - t0) / width)**q
    except ImportError: pass
    from scipy.stats import gennorm
    return gennorm.logpdf(t, q, center, width)
def gennorm_prob(t, q, center=0, width=1):
    '''
    gennorm_prob(t, q) yields the probability density of the generalized error
      distribution, whose PDF is given by q / (2 gamma(1 / q)) exp(-|t|^q).
    genorm_prob(t, q, t0) is equivalent to gennorm_prob(t - t0, q).
    genorm_prob(t, q, t0, s) uses the scale s.
    '''
    (t, q, center, width) = promote(t, q, center, width)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            const = q / (2 * lgamma(1 / q)) / width
            return const * exp(-abs((t - t0) / width)**q)
    except ImportError: pass
    from scipy.stats import gennorm
    return gennorm.pdf(t, q, center, width)
def gumbel_log_prob(t, center=0, lw=1, rw=None):
    '''
    gumbel_log_prob(t) yields the log-probability of the standard Gumbel
      distribution whose PDF is given by exp(-(t + exp(-t))).
    gumbel_log_prob(t, w) yields the log-probability of the standard
      Gumbel distribution with a width of w.
    gumbel_log_prob(t, lw, rw) uses the lw and rw for the widths of the
      left-hand and right-hand side of the Gumbel distribution; i.e.:
      -(t/rw + exp(-t)/lw).
      
    Note that this is not a typical Gumbel distribution definition, but it is a
    very similar distribution nonetheless. Whereas the Gumbel PDF is usually
    defined as being proportional to exp(-((t-t0)/w + exp(-(t-t0)/w))), this
    particular parameterization allows for the linear part of the exponential
    B(T) = t to be scaled separately from the exponential part of the
    exponential A(t) = exp(-t). In a traditional Gumbel distribution,
    the PDF(t) = exp(-[A([t-t0]/w) + B([t-t0]/w)]). In the parameterization
    here, PDF(t) = exp(-[A([t-t0]/a) + B([t-t0]/b)]).
    '''
    if rw is None: rw = lw
    (t, center, lw, rw) = promote(t, center, lw, rw)
    t = (t - center)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            const = torch.log(1.0 / (lw * torch.lgamma(lw/rw)))
            return const - t/rw - torch.exp(-t/lw)
    except ImportError: pass
    const = np.log(1.0 / (lw * lgamma(lw/rw)))
    return const - t/rw - np.exp(-t/lw)
def gumbel_prob(t, center=0, lw=1, rw=None):
    '''
    gumbel_prob(t) yields the probability density of the standard Gumbel
      distribution whose PDF is given by exp(-(t + exp(-t))).
    gumbel_prob(t, w) yields the probability of the standard Gumbel
      distribution with a width of w.
    gumbel_prob(t, lw, rw) uses the lw and rw for the widths of the
      left-hand and right-hand side of the Gumbel distribution; i.e.: the
      probability distribution:
      exp(-(t/rw + exp(-t)/lw)) / (lw^(1/(lw*rw)) gamma(1/(lw*rw))).
      
    Note that this is not a typical Gumbel distribution definition, but it is a
    very similar distribution nonetheless. Whereas the Gumbel PDF is usually
    defined as being proportional to exp(-((t-t0)/w + exp(-(t-t0)/w))), this
    particular parameterization allows for the linear part of the exponential
    B(T) = t to be scaled separately from the exponential part of the
    exponential A(t) = exp(-t). In a traditional Gumbel distribution,
    the PDF(t) = exp(-[A([t-t0]/w) + B([t-t0]/w)]). In the parameterization
    here, PDF(t) = exp(-[A([t-t0]/a) + B([t-t0]/b)]).
    '''
    if rw is None: rw = lw
    (t, center, lw, rw) = promote(t, center, lw, rw)
    t = (t - center)
    try:
        torch = pytorch()
        if torch.is_tensor(t):
            const = 1.0 / (lw * torch.lgamma(lw/rw))
            return const * torch.exp(-t/rw - torch.exp(-t/lw))
    except ImportError: pass
    const = 1.0 / (lw * lgamma(lw/rw))
    return const * np.exp(-t/rw - np.exp(-t/lw))





####################################################################################################
# neuropythy/plans/prfclean.py
# A PIMMS Calcuulation Plan for cleaning retinotopic maps within deduced/labeled visual areas.
# By Noah C. Benson

import numpy as np
import pyrsistent as pyr
import os, sys, gzip, six, types, pimms
from ..util import (is_tuple, is_list)
from .core  import (limit_param, unlimit_param)
from ..math import (totensor, astensor, pytorch, to_torchdtype)
from .. import math


# Initialization of Parameters #####################################################################
@pimms.calc('torch')
def init_pytorch_defaults(dtype='float', device='cpu'):
    '''
    This function provides default values for afferent parameters related to
    PyTorch and checks to ensure that PyTorch is installed.

    Afferent parameters:
     @ dtype May optionally specify the dtype for PyTorch to use in the
       optimization. This may be a PyTorch dtype object or it may be a string,
       in which case it is looked up as a member of the torch module (i.e.,
       'float' 
     @ device The pytorch device name to be used for all tensors (default:
       'cpu').

    Efferent values:
     @ torch The PyTorch module used in the optimization.
    '''
    # Make sure we can import pytorch.
    torch = pytorch()
    # Make sure that we have a valid dtype and device parameters.
    u = totensor(0, dtype=dtype, device=device)
    # If those didn't raise exceptions, it's good enough for now!
    return (torch,)
@pimms.calc('gradients')
def init_gradients(grad_coords=True, grad_cmag_rad=True, grad_cmag_tan=True, grad_ipsi=True):
    '''
    Initializes an empty (mutable!) dictionary for tracking the elements of the
    model that are part of the gradient descent. The gradients dictionary is
    used as a mutable dictionary in which any PyTorch tensor that is part of the
    gradient descent optimization can add itself. These data are needed for the
    PyTorch optimizer.

    Afferent parameters:
     @ grad_coords Specifies whether the tensors representing the coordinates of
       the PRF centers should track their gradients (default: True).
     @ grad_cmag_rad Specifies whether the tensors representing the radial 
       cortical magnification parameters of the retinotopy model should track
       their gradients (default: True).
     @ grad_cmag_tan Specifies whether the tensors representing the tangential
       cortical magnification parameters of the retinotopy model should track
       their gradients (default: True).
     @ grad_ipsi Whether the uvm_ipsi and lvm_ipsi parameters, representing the
       ipsilateral representation in each visual area, are part of the optimized
       parameters whose gradients are required.

    Efferent values:
     @ gradients A mutable dictiionary whose values are PyTorch tensors whose
       gradients are tracked. Any parameter of the optimization should be added
       to this dictionary during the model's initialization.
    '''
    return ({},)
init_params_plan = pimms.plan(gradients=init_gradients,
                              pytorch=init_pytorch_defaults)

# Initialization of Subject Data ###################################################################
@pimms.calc('meas_angle', 'meas_eccen', 'meas_theta', 'meas_prfsize', 'meas_cod',
            'meas_x', 'meas_y', 'meas_properties', 'hemifield_sign')
def init_meas(cortex, device, dtype, retinotopy='prf_'):
    '''
    Calculation:
      (cortex, retinotopy, dtype) => (meas_{angle,eccen,theta,prfsize,vexpl})
    Locates the measured retinotopy data.
    
    Afferent parameters:
     @ cortex The neuropythy cortex/hemisphere object on which the optimization
       occurs.
     @ retinotopy The argument for neuropythy's retinotopy_data function; this
       is typically a string prefix for retinotopy data on the cortex's
       properties ('prf_' for HCP subjects), or a map of the data itself.
    
    Efferent values:
     @ meas_angle The measured polar angle of each vertex on 'cortex' in
       neuropythy's 'visual' coordinate style (positive is clockwise, UVM is
       zero, units are degrees).
     @ meas_theta The measured polar angle of each vertex on 'cortex' in
       neuropythy's 'standard' coordinate style (positive is counter-clockwise,
       RHM is zero, units are radians).
     @ meas_eccen The measured eccentricity of each vertex on 'cotex' in
       visual degrees.
     @ meas_cod The measured variance-explained (COD) of each vertex on
       'cotex' represented as a real number between 0 and 1.
     @ meas_prfsize The measured pRF size of each vertex on 'cotex' in visual
       degrees.
     @ meas_x The measured x coordinate of each vertex.
     @ meas_y The measured y coordinate of each vertex.
     @ meas_properties A persistent map of all the measured properties.
    '''
    import torch, neuropythy as ny
    rd = ny.vision.retinotopy_data(cortex, retinotopy)
    ang = np.array(rd['polar_angle'])
    ecc = np.array(rd['eccentricity'])
    vxp = np.array(rd['variance_explained'])
    rfs = np.array(rd['radius'])
    tht = np.pi/180 * (np.mod(90 - ang + 180, 360) - 180)
    # NaNs need to be converted to 0's
    ang[~np.isfinite(ang)] = 0.0
    ecc[~np.isfinite(ecc)] = 0.0
    vxp[~np.isfinite(vxp)] = 0.0
    rfs[~np.isfinite(rfs)] = 0.0
    tht[~np.isfinite(tht)] = 0.0
    x = np.cos(tht) * ecc
    y = np.sin(tht) * ecc
    logecc = np.log(0.75 + ecc)
    logx = np.cos(tht) * logecc
    logy = np.sin(tht) * logecc
    ps = {'angle':   totensor(ang, dtype=dtype, device=device),
          'eccen':   totensor(ecc, dtype=dtype, device=device),
          'theta':   totensor(tht, dtype=dtype, device=device),
          'prfsize': totensor(rfs, dtype=dtype, device=device),
          'cod':     totensor(vxp, dtype=dtype, device=device),
          'x':       totensor(x, dtype=dtype, device=device),
          'y':       totensor(y, dtype=dtype, device=device)}
    ps = pyr.pmap(ps)
    retval = {('meas_' + k): v for (k,v) in six.iteritems(ps)}
    retval['meas_properties'] = ps
    # hemifiield sign
    retval['hemifield_sign'] = 1 if cortex.chirality == 'lh' else -1
    return retval
@pimms.calc('cortex_labels', 'visual_areas')
def init_labels(cortex, labels, device, torch):
    '''
    Calculation:
      (cortex, labels, label_key) => cortex_labels
    Initializes the visual area labels that are used in the optimization.

    Afferent parameters:
     @ labels Must be the labels used in the optimization. These must be integer
       labels (one per vertex of coftex) or the name of a property of cortex.
       Note that the labels must not use 0 as an ROI: 0 must always indicate
       vertices that should not be optimized.

    Efferent values:
     @ cortex_labels A numpy array of the labels on cortex. 
     @ visual_areas A tuple of the visual area labels that are included in thee
       optimization.
    '''
    if pimms.is_str(labels): labels = cortex.prop(labels)
    if torch.is_tensor(labels): labels = labels.detach().numpy()
    labels = np.array(labels, 'int')
    labels.setflags(write=False)
    if len(labels.shape) != 1 or labels.shape[0] != cortex.vertex_count:
        raise ValueError("labels parameter's length does not match cortex")
    vas = np.setdiff1d(np.unique(labels), [0])
    return (labels, tuple(vas))
@pimms.calc('full_mesh', 'mesh', 'roi_meshes', 'roi_meshindices', 'mesh_labels',
            'label_sort', 'label_indices')
def init_mesh(torch, cortex, meas_properties, cortex_labels, visual_areas):
    '''
    Calculation:
      (cortex, cortex_labels, visual_areas, meas_properties) 
        => (full_mesh, mesh, roi_meshes, labels)
    Loads the midgray mesh from the cortex object and limits it the vertices
    over which the minimization operates.
   
    Afferent parameters:
     @ cortex The neuropythy Cortex object (representing a hemisphere) to use.
    
    Efferent values:
     @ full_mesh The migray mesh for the cortex being modeled.
     @ mesh The midgray surface mesh on which the model is implemented,
       limited to those vertices whose labels is non-zero.
     @ roi_meshes One mesh per ROI in the visual_areas list, each of which is
       the midgray mesh limited to just the vertices in the given ROI. This
       value is a persistent dictionary whose keys are the visual area labels.
     @ roi_meshindices A persistent map whose keys are ROI labels in the mesh
       and whose vales are vectors of the indices in mesh for the vertices of
       the relevant ROIs.
     @ mesh_labels An array of the labels for eeach vertex of the mesh.
     @ label_sort For parameters of the optimization that are potentially part
       of the gradient, it is often desirable to store these in a vector, one
       per visual area. For this reason, we have a label_sort which gives us
       the labels in the order listed in such arrays. See also label_indices.
     @ label_indices For parameters of the optimization that are potentially
       part of the gradient, it is often desirable to store these in a vector,
       one per visual area. For this reason, we have a label_indices element,
       which is a persistent map that gives us the position (as values) when
       a visual area is looked up (as a key). See also label_sort.
    '''
    p = meas_properties.set('visual_areas', cortex_labels)
    p = {k: (v.detach().numpy() if torch.is_tensor(v) else v)
         for (k,v) in six.iteritems(p)}
    cortex = cortex.with_prop(**p)
    full_mesh = cortex.surface('midgray')
    mesh = full_mesh.submesh(cortex_labels != 0)
    roi_meshes = {k: full_mesh.submesh(cortex_labels == k)
                  for k in visual_areas}
    roi_meshes = pyr.pmap(roi_meshes)
    roi_meshindices = pyr.pmap({k: np.where(np.isin(mesh.labels, m.labels))[0]
                                for (k,m) in six.iteritems(roi_meshes)})
    for m in six.itervalues(roi_meshindices): m.setflags(write=False)
    lbls = cortex_labels[mesh.labels]
    lbl_sort = np.sort(np.unique(lbls))
    lbl_indices = pyr.pmap({k:v for (v,k) in enumerate(lbl_sort)})
    return (full_mesh, mesh, roi_meshes, roi_meshindices, lbls, lbl_sort, lbl_indices)
def cp_facearea(ax, bx, cx):
    '''
    cp_faceareas(ax, bx, cx) yields the signed face areas for the triangle
      or triangles with corners ax, bx, and cx (which may be 2 or 3 dimensional
      vectors, or 2 or 3 x N matrices of column-vectors). If the dimensionality
      is 3, the absolute value of the face area is always returned.
    
    Note that ax, bx, and cx must be given in counter-clockwise order, and they
    must all be either numpy arrays or PyTorch tensors.
    '''
    u = bx - ax
    v = cx - ax
    if ax.shape[0] == 2:
        return (u[0]*v[1] - u[1]*v[0]) / 2.0
    else:
        r2 = ((u[0]*v[1] - u[1]*v[0])**2 +
              (u[2]*v[1] - u[1]*v[2])**2 +
              (u[0]*v[2] - u[2]*v[0])**2)
        return math.safesqrt(r2) / 2.0
def vmag_vertex_areas(vmag_summation_matrices, faces, coords):
    '''
    vmag_vertex_areas(matrices, faces,, coords) yields the surface area
      associated with each vertex represented in the coordinate array, coords.
    
    The matrices and faces rguments are expected to be those returned
    by the init_geometry calculation. The coords must be a 2 x N or 3 x N matrix
    where N is the number of vertices. For 3D coordinates, the result is always
    positive; for 2D coordinates, the result is signed depending on the
    fieldsign.
    '''
    torch = pytorch()
    coords = astensor(coords)
    (amtx, bmtx, cmtx) = vmag_summation_matrices
    (a,b,c) = faces
    # The centers of the face and edges.
    (ax, bx, cx) = (coords[:,a], coords[:,b], coords[:,c])
    f0 = (ax + bx + cx) / 3.0
    ab0 = (ax + bx) / 2.0
    bc0 = (bx + cx) / 2.0
    ca0 = (cx + ax) / 2.0
    # Calculate the areas.
    aareas = cp_facearea(ax, ab0, f0) + cp_facearea(ax, f0, ca0)
    bareas = cp_facearea(bx, bc0, f0) + cp_facearea(bx, f0, ab0)
    careas = cp_facearea(cx, ca0, f0) + cp_facearea(cx, f0, bc0)
    # Transfer over to vertces.
    vertex_aareas = torch.sparse.mm(amtx, torch.reshape(aareas, (-1,1)))[:,0]
    vertex_bareas = torch.sparse.mm(bmtx, torch.reshape(bareas, (-1,1)))[:,0]
    vertex_careas = torch.sparse.mm(cmtx, torch.reshape(careas, (-1,1)))[:,0]
    return vertex_aareas + vertex_bareas + vertex_careas
@pimms.calc('faces', 'edges',
            'face_pairs', 'face_pair_edges', 'face_pair_opposites',
            'edge_lengths_mm', 'face_areas_mm2', 'vertex_areas_mm2',
            'vmag_matrices')
def init_geometry(mesh, dtype, device, torch):
    '''
    Calculation:
      (mesh) => (faces, edges, face_pairs, face_pair_opposites, edge_lenghts_mm,
                 face_areas_mm2, vertex_areas_mm2)
    Loads the midgray mesh from the cortex object and limits it using a mask
    formed by the vertices in a spotlight flatmap of the occipital pole.
   
    Efferent values:
     @ faces The 3 x n numpy arrary of the mesh vertex indices of each corner
       of its faces.
     @ edges The 2 x n numpy arrary of vertex indices in the mesh for each edge
       endpoint.
     @ face_pairs The 2 x n numpy array whose columns represent face indices
       are adjacent to each other. The indices of the edges that lie between
       each face pair is given by face_pair_edges.
     @ face_pair_edges The numpy array of edge indices that match the face
       pairs in the 'face_pairs' value. Each pair of faces that share an
       edge is given by the face inidices in face_pairs[k] and the edge 
       between them is given by the edge index in face_pair_edges[k].
     @ face_pair_opposites The 2 x n numpy array of mesh vetex indices of the
       vetices opposite each face in the face_pairs index. For a column k of
       face_pairs, the face indexed by face_pairs[k][0] is across the edge
       indexed by face_pair_edges[k] from the vertex indexed by
       face_pair_opposites[k][0]. The same is true for the face given by 
       ...[k][1].
     @ edge_lengths_mm The lengths of each edge in mesh in mm.
     @ face_areas_mm2 The cotical surface area of each face in mesh in square
       mm.
     @ vertex_areas_mm2 The cortical suface area of each vertex in mesh in
       square mm.
     @ vmag_matrices A 3-tuple of (n x N, n x M, n x M) PyTorch sparse
       matrices, each of whihch corresponds to its matching entry in the
       vmag_indices element. These matrices are designed such that if one
       multiplies them each by a vector of the surface areas (i.e., in the
       visual field) of the faces described in vmag_indices then sums the three
       resulting vectors, thee resulting n-length vector will be the surface
       areas corresponding to each of the n vertices in mesh.
    '''
    dtype = to_torchdtype(dtype)
    edges = mesh.tess.indexed_edges.T
    faces = mesh.tess.indexed_faces.T
    # Figure out the face-pairs.
    (fp1,fp2,fp_edges) = np.transpose(
        [(k[0],k[1],ii) for (ii,k) in enumerate(mesh.tess.edge_faces) if len(k) == 2])
    fps = np.array([fp1,fp2])
    face_pairs = fps.T
    f1other = np.squeeze([faces[f1, ~np.isin(faces[f1], faces[f2])]
                          for (f1,f2) in face_pairs])
    f2other = np.squeeze([faces[f2, ~np.isin(faces[f2], faces[f1])]
                          for (f1,f2) in face_pairs])
    fp_opps = np.array([f1other, f2other])
    fps = face_pairs.T
    # Get the edge lengths and triangle surface areas from the midgray surface.
    els = totensor(mesh.edge_lengths, dtype=dtype, device=device)
    fas = totensor(mesh.face_areas, dtype=dtype, device=device)
    # Calculate the vmag summation matrix.
    
    # The best way to do this is to split each face up into six faces. Each corner gets two of
    # these subfaces.
    (a,b,c) = mesh.tess.indexed_faces
    n = mesh.tess.face_count
    (amtx, bmtx, cmtx) = [
        torch.sparse_coo_tensor(
            torch.tensor(np.array([corner, np.arange(n)]), dtype=torch.long),
            torch.ones(n),
            size=(mesh.vertex_count, n),
            dtype=dtype,
            device=device)
        for corner in (a,b,c)]
    matrices = (amtx, bmtx, cmtx)
    # Go ahead and use these to return the vertex surface areas.
    faces = totensor(np.array(faces.T), dtype=torch.long)
    edges = totensor(np.array(edges.T), dtype=torch.long)
    vas = vmag_vertex_areas(matrices, faces, mesh.coordinates)
    vas = totensor(vas, dtype=dtype, device=device)
    # Return everything.
    return (faces, edges, fps, fp_edges, fp_opps, els, fas, vas, matrices)
init_geometry_plan = pimms.plan(geometry=init_geometry)
@pimms.calc('roi_geometries', 'roi_cortical_surface_areas', 'boundary_edges')
def init_geometries(mesh, mesh_labels, roi_meshes, label_sort, dtype, device, torch):
    '''
    Calculates an init_geometry() imap for each roi-mesh. See init_geometry for
    more information.

    Efferent values:
      @ roi_geometries A persistent map of geometry data for each ROI in the
        optimization. See init_geometry for more information.
      @ roi_cortical_surface_areas A PyTorch tonsor, in label-sort order, of the
        cortical surface areas of each ROI in square mm.
      @ boundary_edges A persistent map whose keys are (area1, area2) tuples
        containing two adjacent visual area labels and whose keys are 2 x N
        numpy arrays (u, v) where u and v are vectors of vertex indices for
        edges that cross the relevant boundary.
    '''
    r = {}
    sas = []
    for (k,msh) in six.iteritems(roi_meshes):
        r[k] = init_geometry_plan(mesh=msh, dtype=dtype, device=device, torch=torch)
        sas.append(np.sum(msh.face_areas))
    sas = totensor(np.array(sas), dtype=dtype, device=device)
    # Make the edge boundaries
    (u,v) = mesh.tess.indexed_edges
    (lu, lv) = (mesh_labels[u], mesh_labels[v])
    ii = np.where((lu == 1) & (lv == 3))[0]
    jj = np.where((lu == 3) & (lv == 1))[0]
    # We only care about edges without that cross label boundaries.
    ii = lu != lv
    (u,v,lu,lv) = (u[ii], v[ii], lu[ii], lv[ii])
    # We also want the lower vertex label to come first:
    ii = lu > lv
    tmp = np.array(lu[ii])
    lu[ii] = lv[ii]
    lv[ii] = tmp
    tmp = np.array(u[ii])
    u[ii] = v[ii]
    v[ii] = tmp
    ebs = {}
    for k1 in label_sort:
        for k2 in label_sort:
            if k1 >= k2: continue
            ii = (lu == k1) & (lv == k2)
            tmp = np.array([u[ii], v[ii]])
            if tmp.shape[1] == 0: continue
            tmp.setflags(write=False)
            ebs[(k1, k2)] = tmp
    # We also want edges with the outer mesh; we store these as (u,u) because they have no
    # neighbor in the mesh.
    outer_vs = [xx
                for (ii,k) in enumerate(mesh.tess.edge_faces)
                if len(k) == 1
                for xx in mesh.tess.indexed_edges[:,ii]]
    outer_vs = np.unique(outer_vs)
    outer_ls = mesh_labels[outer_vs]
    for lbl in label_sort:
        ii = outer_ls == lbl
        uu = outer_vs[ii]
        if len(uu) == 0: continue
        tmp = np.array([uu, uu])
        tmp.setflags(write=False)
        ebs[(0,lbl)] = tmp
    return (pyr.pmap(r), sas, pyr.pmap(ebs))
init_subject_plan = pimms.plan(
    {'meas':       init_meas,
     'labels':     init_labels,
     'mesh':       init_mesh,
     'geometries': init_geometries})

# Ipsilateral Parameters ###########################################################################
@pimms.calc('uvm_ipsi_tan', 'lvm_ipsi_tan')
def init_ipsi_tan_params(gradients, dtype, device, grad_ipsi, label_sort, torch,
                         prior_uvm_ipsi=0.09, prior_lvm_ipsi=0.17):
    '''
    Initializes the "free" versions of the ipsilateral representation parameters
    uvm_ipisi_tan and lvm_ipsi_tan. Both are the radians of ipsilateral
    representation from the respective boundary (i.e., positive values are
    always the number of radians into the ipsilateral visual field that is
    represented in the retinotopic map. The free values are in fact the tangent
    of the values used in the model, preventing very extreme values or values
    beyond +/- pi/2 at all. Note that negative values indicate that the data do
    not cover the entire contralateral hemifield.

    Afferent parameters:
     @ prior_uvm_ipsi The prior value of the upper-vertical-meridian's
       ipsilateral representation. This value is the number of radians into the
       ipsilateral visual field that the PRF representations of all visual areas
       extend for the upper vertical meridian.
     @ prior_lvm_ipsi The prior value of the lower-vertical-meridian's
       ipsilateral representation. This value is the number of radians into the
       ipsilateral visual field that the PRF representations of all visual areas
       extend for the lower vertical meridian.

    Efferent values:
     @ uvm_ipsi_tan The tangent of the ipsilateral relresentattion angle for the
       upper visual field.
     @ lvm_ipsi_tan The tangent of the ipsilateral relresentattion angle for the
       lower visual field.
    '''
    uvm = astensor(0.0 if prior_uvm_ipsi is None else prior_uvm_ipsi)
    lvm = astensor(0.0 if prior_lvm_ipsi is None else prior_lvm_ipsi)
    if uvm.shape == (): uvm = uvm * torch.ones(len(label_sort))
    if lvm.shape == (): lvm = lvm * torch.ones(len(label_sort))
    for u in uvm:
        if u > np.pi/2 or u < -np.pi/2:
            raise ValueError("prior UVM ipsilateral angle must be between + and - pi/2")
    for l in lvm:
        if l > np.pi/2 or l < -np.pi/2:
            raise ValueError("prior LVM ipsilateral angle must be between + and - pi/2")
    uvm = torch.tan(uvm)
    lvm = torch.tan(lvm)
    uvm = totensor(uvm, dtype=dtype, device=device, requires_grad=grad_ipsi)
    lvm = totensor(lvm, dtype=dtype, device=device, requires_grad=grad_ipsi)
    if grad_ipsi:
        gradients['uvm_ipsi_tan'] = uvm
        gradients['lvm_ipsi_tan'] = lvm
    return (uvm, lvm)
@pimms.calc('uvm_ipsi', 'lvm_ipsi')
def calc_ipsi_params(uvm_ipsi_tan, lvm_ipsi_tan, torch):
    '''
    Calculates the ipsilateral parameters from the tan-parameters (i.e. converts
    the "free" parameters to limited parameters between +/- pi/2).

    Efferent values:
     @ uvm_ipsi The angle of extent into the ipsilateral hemifield of the
       representation of the upper vertical meridian in the model.
     @ lvm_ipsi The angle of extent into the ipsilateral hemifield of the
       representation of the lower vertical meridian in the model.
    '''
    return (torch.atan(uvm_ipsi_tan), torch.atan(lvm_ipsi_tan))
init_ipsi_plan = pimms.plan(ipsi_ten_params=init_ipsi_tan_params,
                            ipsi_params=calc_ipsi_params)


# Optimized Visual Field Coordinates ###############################################################
@pimms.calc('theta', 'eccen')
def init_visual_coords(meas_theta, meas_eccen, mesh, cortex, gradients, grad_coords, dtype, device,
                       start_coords=None):
    '''
    Initializes the visual field coordinates that get minimized as part of the
    optimization scheme.

    Afferent parameters:
      @ start_coords May specify the (2 x N) coordinate matrix of measurements
        to use as the starting point in the optimization. The N must be the
        number of vertices in the cortex; though values are only needed wherever
        there are non-zero labels. If None, then the measurements are used as
        the starting coordinates. The first row of the start coordinates must be
        the polar angle theta given in counter-clockwise radians starting from
        the right horizontal meridian, and the second row must be the
        eccentricity of the vertices in visual degrees.

    Efferent values:
      @ theta The polar angle values over which the optimization takes place.
      @ eccen The eccentricity values over which the optimizationn takes place.
    '''
    if start_coords is None:
        # Use the measurements
        tht = np.array(meas_theta.detach().numpy()[mesh.labels])
        ecc = np.array(meas_eccen.detach().numpy()[mesh.labels])
    else:
        viscoords = math.asarray(start_coords)
        if len(viscoords.shape) != 2: raise ValueError('start_coords must be a 2 x N matrix')
        (rs,cs) = viscoords.shape
        if rs != 2 and rs == cortex.vertex_count and cs == 2:
            viscoords = viscoords.T
        elif rs != 2 or cs != cortex.vertex_count:
            raise ValueError('start_coords must be 2 x N where N is number of cortex vetices')
        (tht,ecc) = [np.array(u) for u in viscoords[:, mesh.labels]]
    # Make this into a PyTorch tensor.
    tht = totensor(tht, dtype=dtype, device=device, requires_grad=grad_coords)
    ecc = totensor(ecc, dtype=dtype, device=device, requires_grad=grad_coords)
    # Add this tensor to the gradients if it's one of them.
    if grad_coords: 
        gradients['theta'] = tht
        gradients['eccen'] = ecc
    return (tht, ecc)
@pimms.calc('x', 'y')
def calc_cartesian(theta, eccen, torch):
    '''
    Converts polar angle and eccentricity into x and y.

    Efferent values:
      @ x The x coordinate of each vertex in the visual field.
      @ y The y coordinate of each vertex in the visual field.
    '''
    return (torch.cos(theta) * eccen, torch.sin(theta) * eccen)
@pimms.calc('logeccen', 'logx', 'logy')
def calc_logcartesian(theta, eccen, torch):
    '''
    Converts polar angle and eccentricity into log-x and log-y.

    Efferent values:
      @ logx The x coordinate of each vertex in the visual field, scaled such
        that the eccentricity of each point is log(0.75 + eccen) - log(0.75).
      @ logy The y coordinate of each vertex in the visual field, scaled such
        that the eccentricity of each point is log(0.75 + eccen) - log(0.75).
    '''
    logecc = torch.log(0.75 + eccen) - log(0.75)
    return (torch.cos(theta) * logeccen, torch.sin(theta) * logeccen)
coords_plan = pimms.plan(visual_coords=init_visual_coords,
                         #logcartesian=calc_logcartesian,
                         cartesian=calc_cartesian)

# Cortical Magnification ###########################################################################
@pimms.calc('face_areas_deg2', 'vmag_faces')
def calc_vmag_faces(faces, face_areas_mm2, x, y):
    '''
    Calculation:
      (faces, edges, face_areas_mm2, x, y)
      ==> (face_areas_deg2, vmag_faces)
    
    Efferent values:
      @ face_areas_deg2 The surface area of each face in the visual field.
      @ vmag_faces The visual magnification of each face.
    '''
    (a,b,c) = faces
    (xa, ya) = (x[a], y[a])
    xab = x[b] - xa 
    yab = y[b] - ya
    xac = x[c] - xa 
    yac = y[c] - ya
    fareas = (xab*yac - yab*xac) / 2.0
    fvmags = fareas / face_areas_mm2
    return (fareas, fvmags)
@pimms.calc('vertex_areas_deg2', 'vmag')
def calc_vmag_vertices(faces, vertex_areas_mm2, vmag_matrices, x, y, torch):
    '''
    Calculation:
      (faces, vertex_areas_mm2, vmag_matrices, x, y)
      ==> (vertex_areas_deg2, vmag)
    
    Efferent values:
      @ vertex_areas_deg2 The surface area of each vertex in the visual field.
      @ vmag The visual magnification of each vertex.
    '''
    # Calculate the vertex areas.
    coords = torch.stack([x,y])
    vareas = vmag_vertex_areas(vmag_matrices, faces, coords)
    # The rest is easy.
    return (vareas, vareas / vertex_areas_mm2)
vmag_mesh_plan = pimms.plan(vmag_vertices=calc_vmag_vertices, vmag_faces=calc_vmag_faces)
@pimms.calc('roi_vmags')
def calc_roi_vmags(roi_meshindices, roi_geometries, x, y, torch):
    '''
    Calculates an imap of vmag data for each roi in the optimization.

    Efferent values:
      @ roi_vmags A persistent map whose keys are ROI labels and whose values
        are imaps obtained from running the vmag_mesh_plan on each ROI mesh.
    '''
    from neuropythy.util import curry
    runplan = lambda k: vmag_mesh_plan(
        faces=roi_geometries[k]['faces'],
        edges=roi_geometries[k]['edges'],
        face_areas_mm2=roi_geometries[k]['face_areas_mm2'],
        vertex_areas_mm2=roi_geometries[k]['vertex_areas_mm2'],
        vmag_matrices=roi_geometries[k]['vmag_matrices'],
        x=x[roi_meshindices[k].copy()],
        y=y[roi_meshindices[k].copy()],
        torch=torch)
    res = pimms.lmap({k: curry(runplan, k) for k in roi_geometries.keys()})
    return (res,)
vmag_plan = pimms.plan(roi_vmags=calc_roi_vmags)

# Cortical Magnification Model #####################################################################
def cmmdl_tanmult_rescale(theta, hemifield_sign, ui=0.09, li=0.17):
    '''
    Yields the theta parameter rescaled such that the upper vertical field ends
    ui radians into the ipsilateral visual field and such that the lower vertical
    field ends li radians into the ipsilateral visual field. The returned value
    is scaled such that a value of pi/2 + ui in the input theta has a value pi/2
    in the output theta, and equivalently with -pi/2 and li in the lower visual
    field. The x-value sign of the contralateral visual field is the second
    argument to this function (and the signs in the above explanation are
    reversed when this argument is -1).
    '''
    torch = pytorch()
    (hpi, tau) = (math.half_pi, math.tau)
    if hemifield_sign == -1: theta = torch.remainder(tau - theta, tau) - hpi
    th = torch.zeros(theta.shape)
    gt = theta > 0
    lt = ~gt
    th[gt] = theta[gt] / (hpi + ui) * hpi
    th[lt] = theta[lt] / (hpi + li) * hpi
    if hemifield_sign == -1: th = torch.remainder(tau - th, tau) - hpi
    return th
def cmmdl_tanmult_sinusoid(theta, hva=0.5, vma=0.5):
    '''
    Yields the angular cortical magnification multiplier based on the polar
      angle theta, which is measured in counter-clockwise radians from the RHM.
    
    The sinusoid version of this function yields a polar angle multiplier based
    on the sum of sinusoids.
    '''
    torch = pytorch()
    hvpart = hva * torch.cos(2 * theta)
    thsin  = torch.sin(theta)
    ulpart = vma * torch.sign(thsin) * thsin**2
    return 1.0 + 0.5*(hvpart - ulpart)
def cmmdl_tanmult_beta(theta, a=1.0, b=1.0):
    '''
    Yields the angular cortical magnification multiplier based on the polar
      angle theta, which is measured in counter-clockwise radians from the RHM.
    
    The beta-distribution version of this function uses a beta distribution, 
    spread across -pi to pi.
    '''
    torch = pytorch()
    thuvm = torch.remainder(math.half_pi - theta + np.pi, math.tau) - np.pi
    thuvm = torch.abs(thuvm)
    theta = torch.remainder(math.half_pi - thuvm + np.pi, math.tau) - np.pi
    theta = (theta + math.half_pi) / np.pi
    beta = torch.distributions.Beta(astensor(a), astensor(b))
    return torch.exp(beta.log_prob(theta))
def cmmdl_hhcmag(eccen, c1=17.3, c2=0.75):
    '''
    cmmdl_hhcmag(eccen) yields the linear radial cortical magnification using
      the Horton and Hoyt (1991) formula.
    '''
    return c1 / (c2 + eccen)
def cmmdl_hhcmag2(eccen, c1=17.3, c2=0.75):
    '''
    cmmdl_hhcmag2(eccen) yields the areal radial cortical magnification using
      the Horton and Hoyt (1991) formula.
    '''
    return cmmdl_hhcmag(eccen, c1=c1, c2=c2)**2
def cmmdl_cmag2_beta(theta, eccen,
                     c1=17.3, c2=0.75, a=1.0, b=1.0):
    '''
    Yields the areal cortical magnification prediction for the given polar
      angle and eccentricity using both the Horton and Hoyt (1991) equation
      as well as the tangential cortical magnification multiplier, based on a
      beta distribution.
    '''
    rmag = cmmdl_hhcmag2(eccen, c1=c1, c2=c2)
    tmlt = cmmdl_tanmult_beta(theta, a=a, b=b)
    return rmag * tmlt
def cmmdl_cmag2_sinusoid(theta, eccen,
                         c1=17.3, c2=0.75, hva=0.5, vma=0.5):
    '''
    Yields the areal cortical magnification prediction for the given polar
      angle and eccentricity using both the Horton and Hoyt (1991) equation
      as well as the tangential cortical magnification multiplier, base on a sum
      of sinusoids.
    '''
    rmag = cmmdl_hhcmag2(eccen, c1=c1, c2=c2)
    tmlt = cmmdl_tanmult_sinusoid(theta, hva=hva, vma=vma)
    return rmag * tmlt
def cmmdl_tan_inparams(method, params):
    torch = pytorch()
    if method is None:
        return params
    elif method == 'sinusoid':
        return torch.atan(astensor(params)) * 2 / np.pi
    elif method == 'beta':
        return torch.exp(torch.atan(astensor(params)) * 3 / math.half_pi)
    else:
        raise ValueError("unknown cmag2 tangential method: %s" % (method,))
def cmmdl_tan_defaultparams(method):
    if method is None:
        return None
    elif method == 'sinusoid':
        return (0.5, 0.5)
    elif method == 'beta':
        return (1.0, 1.0)
    else:
        raise ValueError("unknown cmag2 tangential method: %s" % (method,))
def cmmdl_tan_outparams(method, params):
    torch = pytorch()
    if method is None:
        return params
    elif method == 'sinusoid':
        return torch.tan(astensor(params) * math.half_pi)
    elif method == 'beta':
        return torch.tan(torch.log(astensor(params)) / 3 * math.half_pi)
    else:
        raise ValueError("unknown cmag2 tangential method: %s" % (method,))
def cmmdl_cmag2(theta, eccen, hemifield_sign, method='sinusoid',
                params=[0.0, 0.0], c1=17.3, c2=0.75, ui=0.09, li=0.17):
    '''
    Yields the areal cortical magnification using either cmmdl_cmag2_sinusoid,
    with parameters (hva, vma) and method name 'sinusoid', or cmmdl_cmag2_beta,
    with parameters (a, b). 
    '''
    #th = cmmdl_tanmult_rescale(theta, hemifield_sign, ui=ui, li=li)
    th = theta
    if method is None:
        return cmmdl_hhcmag2(eccen, c1=c1, c2=c2)
    elif method == 'sinusoid':
        (hva, vma) = params
        return cmmdl_cmag2_sinusoid(th, eccen, c1=c1, c2=c2, hva=hva, vma=vma)
    elif method == 'beta':
        torch = pytorch()
        (a, b) = params
        return cmmdl_cmag2_beta(th, eccen, c1=c1, c2=c2, a=a, b=b)
    else:
        raise ValueError("unknown cmag2 tangential method: %s" % (method,))
def cmmdl_c1(area, max_eccen, c2):
    '''
    Yields the value of c1 in the cmmdl that is appropriate for a visual area
    with the given surface-area, within the given max_eccen central degrees,
    and the parameter c2.
    '''
    torch = pytorch()
    c2_maxecc = c2 + max_eccen
    den = np.pi * (torch.log(c2_maxecc / c2) - max_eccen / c2_maxecc)
    return math.safesqrt(area / den) # #TODO is the safesqrt necessary here?
@pimms.calc('cmag_eccen_logoffsets', 'fieldsigns_tensor', 'cmag_tan_params_free')
def init_cmag_params(dtype, device, gradients, torch, label_sort, grad_cmag_rad, grad_cmag_tan,
                     fieldsigns=Ellipsis,
                     cmag_tan_method='sinusoid',
                     prior_cmag_eccen_offsets=0.75,
                     prior_cmag_tan_params=None):
    '''
    Initializes the model parameters required to calculate the cortical
    magnification of each vertex and/or visual field position. This
    calculation initializes the "free" versions of the cmag_eccen_offsets
    parameters--i.e., the version that can range from -inf to inf but that gets
    limited into a particular range for the actual model calculation.
    
    Afferent parameters:
     @ prior_cmag_eccen_offsets The initial value of the offset parameter c2 
       from Horton and Hoyt's (1991) equation cmag(ecc) = c1 / (c2 + ecc).
       This should be either a single number, which is used for all visual
       areas, or a dictionary of visual area labels mapped to values.
     @ fieldsigns The fieldsign values for each visual area. This may either be
       a dictionary whose keys are visual area labels and whose values are all
       either 1 or -1, or it can be Ellipsis or None. If the value is Ellipsis
       (the default), then the dictionary visual_area_field_signs from the
       neuropythy.vision package is used. If the value is None, then even
       labels are given fieldsigns of 1 and odd labels are given fieldsigns of
       -1.
     @ cmag_tan_method The method to use for calculating tangential cortical
       magnification. This may be either 'sinusoid' (the default), 'beta', or
       None, in which case tangential cortical magnification is modeled to be
       uniform across polar angle. For thee sinusoid, the parameters are tan_hva
       and tan_vma where the actual hva and vma are arctan(tan_hva)*2/pi and
       arctan(tan_vma)*2/pi, limiting them both to the +/- 1 range. For the
       'beta' method, the params are loga and logb, where the beta-distribution
       parameters that are used are simply a = exp(loga) and b = exp(logb).
     @ prior_cmag_tan_params The initial values of the tangential cortical
       magnificatiioin model parametters, whose interpretation depends on the
       value given for cmag_tan_method (see it for more details). Note that
       cmag_tan_method describes parameters such as loga, which are used in the
       model via the transformation a = exp(loga). The value that is given for
       the prior should *not* be the transformed param--i.e., provide a not loga
       as the prioir value.

    Efferent values:
     @ cmag_eccen_logoffsets The log of the cortical magnification offset
       parameter. As a free parameter, the value may be any real number; the
       version used in the cortical magnification model is the exp() of this
       value.
     @ fieldsigns_tensor A tensor of the fieldsigns of each visual area in the
       order provided by label_sort. Each of these values must be either 1 or
       -1.
     @ cmag_tan_params_free The free (untransformed) matrix of cmag tangential
       parameters. The matrix is N x D where N is the number of visual areas and
       D is the number of parameters (typically 2).
    '''
    # The c2 parameter.
    dtype = to_torchdtype(dtype)
    pri = prior_cmag_eccen_offsets
    log_de = totensor(np.zeros(len(label_sort)),
                      dtype=dtype, device=device, requires_grad=False)
    if pimms.is_map(pri):
        for (ii,k) in enumerate(label_sort):
            log_de[ii] = pri[k]
    elif torch.is_tensor(pri) or pimms.is_number(pri):
        log_de[:] = pri
    else:
        raise ValueError("Could not understand prior_cmag_eccen_offsets param")
    for x in log_de:
       if not torch.isfinite(x) or x <= 0:
           raise ValueError('prior_cmag_eccen_offsets values must be >= 0')
    log_de = torch.log(log_de)
    if grad_cmag_rad:
        log_de = log_de.clone().detach().requires_grad_(True)
        gradients['cmag_eccen_logoffsets'] = log_de
    # Fieldsigns.
    if fieldsigns is Ellipsis:
        from neuropythy.vision import visual_area_field_signs
        fieldsigns = visual_area_field_signs
    elif fieldsigns is None:
        fieldsigns = {k: (-1 if (k % 2) == 1 else 1) for k in label_sort}
    fieldsigns = [fieldsigns[k] for k in label_sort]
    for fs in fieldsigns:
        if fs != 1 and fs != -1:
            raise ValueError('fieldsign values must be either 1 or -1')
    fieldsigns = totensor(fieldsigns, dtype=dtype, device=device)
    # The tangential magnification parameters.
    if cmag_tan_method is None:
        tanparams = None
    else:
        if cmag_tan_method == 'beta':
            import warnings
            warnings.warn("cmag_tan_method='beta' is unfinished and frequently results in NaNs")
        tanparams = torch.zeros((len(label_sort), 2), dtype=dtype, device=device)
        pri = prior_cmag_tan_params
        if pri is None: pri = cmmdl_tan_defaultparams(cmag_tan_method)
        if pimms.is_map(pri):
            for (ii,k) in enumerate(label_sort):
                tanparams[ii,:] = pri[k]
        else:
            tanparams[:] = astensor(pri)
        tanparams = cmmdl_tan_outparams(cmag_tan_method, tanparams)
        if grad_cmag_tan:
            tanparams = tanparams.clone().detach().requires_grad_(True)
            gradients['cmag_tan_params_free'] = tanparams
    # Return.
    return (log_de, fieldsigns, tanparams)
@pimms.calc('cmag_eccen_offsets', 'cmag_tan_params')
def calc_cmag_params(torch, cmag_eccen_logoffsets, cmag_tan_method, cmag_tan_params_free):
    '''
    Calculates the model parameters required to calculate the cortical
    magnification of each vertex and/or visual field position. This
    calculation involves running the free versions of each pameter through the
    arctangent function in order to convert an infinite range to the range
    (0,1) then linearly expanding that to the parameter's limit.
    
    Efferent values:
     @ cmag_eccen_offsets The offset c2 in Horton and Hoyt's 1991 cortical
       magnification equation cmag(eccen) = c1/(c2 + eccen). This value is
       derived from the cmag_eccen_logoffset, which is the actual parameter
       over which minimization occurs. This is because any real value of the
       log-offset results in a valid value for the offset, preventing the offset
       from derailing the search.
     @ cmag_tan_params The "unfree" version of the cmag_tan_params_free. This is
       a matrix PyTorch-tensor of values where each row is a set of parameters
       for the tangential cortical magnification model (see cmag_tan_method for
       addititonal information).
    '''
    return (torch.exp(cmag_eccen_logoffsets),
            cmmdl_tan_inparams(cmag_tan_method, cmag_tan_params_free))
@pimms.calc('cmag_scale')
def calc_cmag_scale(eccen, label_sort, cmag_eccen_offsets,
                    max_eccen, roi_cortical_surface_areas, dtype):
    '''
    Calculates the 'cmag_scale' meta-parameter for each visual area.

    Afferent parameters:
     @ max_eccen The maximum eccentricity in the optimized maps. This should be
       the ecceentricity of the peripheral boundary of the labeled maps.
    
    Efferent values:
     @ cmag_scale The meta-parameter (c1) that can be calculated from the
       'cmag_eccen_offsets' (c2) parameter and the total visual area size,
       based on the fact that the cortical magnification used in the model
       conforms to Horton and Hoyt's (1991) equation:
       cmag2(ecc) = (c1 / (c2 + ecc))^2.
       The calculation of cmag_scale uses the surface area of the region of
       each visual area that is within max_stim_eccen to calculate the
       parameter value.
    '''
    r = cmmdl_c1(roi_cortical_surface_areas, max_eccen, cmag_eccen_offsets)
    return (r,)
@pimms.calc('model_vmags')
def calc_model_vmags(theta, eccen, fieldsigns_tensor, hemifield_sign,
                     cmag_eccen_offsets, cmag_scale, cmag_tan_method, cmag_tan_params,
                     uvm_ipsi, lvm_ipsi, roi_meshindices, label_sort,
                     dtype, device, torch):
    '''
    Calculates the per-visual-area visual magnification values as predicted by
    the cortical magnification model.
      
    Efferent values:
      @ model_vmags A persistent map whose keys are visual area labels and whose
        values are tensors of the visual magnification values predicted at the
        vertices of each ROI-mesh.
    '''
    c1s = cmag_scale
    c2s = cmag_eccen_offsets
    tps = cmag_tan_params
    vmags = {}
    for (ii,k) in enumerate(label_sort):
        roi_ii = np.array(roi_meshindices[k])
        tht = theta[roi_ii]
        ecc = eccen[roi_ii]
        cm = cmmdl_cmag2(tht, ecc, hemifield_sign,
                         c1=c1s[ii], c2=c2s[ii], params=tps[ii],
                         method=cmag_tan_method, ui=uvm_ipsi[ii], li=lvm_ipsi[ii])
        vmags[k] = fieldsigns_tensor[ii] / cm
    return (pyr.pmap(vmags),)
@pimms.calc('model_vmag_faces')
def calc_model_vmag_faces(theta, eccen, fieldsigns_tensor, hemifield_sign,
                          cmag_eccen_offsets, cmag_scale, cmag_tan_method, cmag_tan_params,
                          uvm_ipsi, lvm_ipsi, roi_meshindices, roi_meshes,
                          label_sort, dtype, device, torch):
    '''
    Calculates the per-visual-area visual magnification values as predicted by
    the cortical magnification model for the faces in the mesh.
      
    Efferent values:
      @ model_vmag_faces A persistent map whose keys are visual area labels and
        whose values are tensors of the visual magnification values predicted at
        the triangles of each ROI-mesh.
    '''
    c1s = cmag_scale
    c2s = cmag_eccen_offsets
    tps = cmag_tan_params
    vmags = {}
    for (ii,k) in enumerate(label_sort):
        roi_ii = np.array(roi_meshindices[k])
        tht = theta[roi_ii]
        ecc = eccen[roi_ii]
        (a,b,c) = np.array(roi_meshes[k].tess.indexed_faces)
        tht = (tht[a] + tht[b] + tht[c]) / 3.0
        ecc = (ecc[a] + ecc[b] + ecc[c]) / 3.0
        cm = cmmdl_cmag2(tht, ecc, hemifield_sign,
                         c1=c1s[ii], c2=c2s[ii], params=tps[ii],
                         method=cmag_tan_method, ui=uvm_ipsi, li=lvm_ipsi)
        vmags[k] = fieldsigns_tensor[ii] / cm
    return (pyr.pmap(vmags),)
model_plan = pimms.plan(init_cmag_params=init_cmag_params,
                        cmag_params=calc_cmag_params,
                        cmag_scale=calc_cmag_scale,
                        model_vmags=calc_model_vmags,
                        model_vmag_faces=calc_model_vmag_faces)

# Model Loss #######################################################################################
@pimms.calc('meas_stddev', 'meas_log_2stddev')
def calc_meas_stddev(torch, meas_eccen, meas_cod, meas_weight=None):
    '''
    Calculates the standard deviation of the measurement distributions that are
    used to calculate the likelihood of the measurements given the data.

    The standard deviation formula for a vertex u is as follows:
      stddev(u) = (0.75 + meas_eccen) / (0.25 + 3.75 * min_cod) / weight

    Afferent parameters:
     @ meas_weight A vector of weights, one per vertex in mesh, that specifies
       how strongly the model should believe that particular vertex's weight.
       Weights are the divisor for the meas_stddev. See calc_meas_stddev for
       more information about the standard deviation formula.
    '''
    if meas_weight is None: meas_weight = 1.0
    stddev = (0.75 + meas_eccen) / (0.25 + 3.75*meas_cod) / meas_weight
    return (stddev, torch.log(2.0 * stddev))
@pimms.calc('meas_likelihood', 'meas_likelihoods')
def calc_meas_likelihood(mesh, meas_x, meas_y, x, y, meas_stddev, meas_log_2stddev, torch):
    '''
    Calculates the measurement likelihood, given the retinotopy model, of the
    data. This is the sum over vertices of the log probabilities of a laplacian
    noise distribution around the predicted value. The standard devations of
    these distributions is meas_stddev.

    Efferent values:
     @ meas_likelihood The log likelihood of the PRF measurements given the PRF
       model.
     @ meas_likelihoods A PyTorch vector of log likelihood values, one per mesh
       vertex, of the measurements given the model.
    '''
    ii = mesh.labels.copy()
    meas_x = meas_x[ii]
    meas_y = meas_y[ii]
    meas_stddev = meas_stddev[ii]
    meas_log_2stddev = meas_log_2stddev[ii]
    dist = math.safesqrt((meas_x - x)**2 + (meas_y - y)**2)
    llhoods = -dist/meas_stddev # - meas_log_2stddev
    return (torch.sum(llhoods), llhoods)
@pimms.calc('vmag_likelihood', 'vmag_likelihoods')
def calc_vmag_likelihood(roi_vmags, model_vmag_faces, label_sort, 
                         dtype, device, torch):
    '''
    Calculates the log likelihood of the visual magnification model.

    Efferent values:
     @ vmag_likelihood The log-likelihood of the vmag component of the model.
     @ vmag_likelihoods A persistent map whose keys are visual area labels and
       whose values are PyTorch vectors of the log likelihood
    '''
    r = {}
    tot = totensor(0.0, dtype=dtype, device=device)
    for k in label_sort:
        #coord_vm = torch.log(0.5 + roi_vmags[k]['vmag'])
        #model_vm = torch.log(0.5 + model_vmags[k])
        coord_vm = roi_vmags[k]['vmag_faces']
        model_vm = model_vmag_faces[k]
        llhood = -(coord_vm - model_vm)**2
        r[k] = llhood
        tot = tot + torch.sum(llhood)
    return (tot, pyr.pmap(r))
visual_area_boundary_fns = pyr.m(
    VM=       lambda t,e,x,y,ui,li,mxec: math.branch(pytorch().sin(t) > 0, t - ui, t - li),
    HM=       lambda t,e,x,y,ui,li,mxec: y,
    periphery=lambda t,e,x,y,ui,li,mxec: e - mxec,
    fovea=    lambda t,e,x,y,ui,li,mxec: e)
visual_area_boundaries = pyr.pmap(
    {(1, 2): 'VM',
     (0, 1): 'periphery',
     (2, 3): 'HM',
     (0, 2): 'periphery',
     (0, 3): ('periphery', 'VM')})
def boundary_distances(lbl1, lbl2, theta, eccen, x, y, uvm_ipsi, lvm_ipsi, max_eccen,
                       boundaries=visual_area_boundaries,
                       boundary_fns=visual_area_boundary_fns):
    '''
    Calculates the square of the boundary distances for the given values.
    '''
    torch = pytorch()
    vab = boundaries.get((lbl1, lbl2), None)
    if vab is None: return None
    if is_tuple(vab):
        if len(vab) == 0: return None
        r = [boundary_fns[subvab](theta, eccen, x, y, uvm_ipsi, lvm_ipsi, max_eccen)**2
             for subvab in vab]
        return torch.min(*r)
    else:
        return boundary_fns[vab](theta, eccen, x, y, uvm_ipsi, lvm_ipsi, max_eccen)**2
@pimms.calc('boundary_likelihood', 'boundary_likelihoods')
def calc_boundary_likelihood(x, y, theta, eccen, boundary_edges, uvm_ipsi, lvm_ipsi, max_eccen,
                             hemifield_sign, label_sort, torch):
    '''
    Calculates the likelihood of the visual area boundaries given the model.

    Efferent values:
     @ boundary_likelihood The total log likelihood of all the boundary edges.
     @ boundary_likelihoods The log likelihood of each of the boundary edges.
    '''
    uvm_ipsi = math.half_pi + hemifield_sign*uvm_ipsi
    lvm_ipsi = -math.half_pi - hemifield_sign*lvm_ipsi
    llhs = {}
    llhtot = 0.0
    for (k, uv) in six.iteritems(boundary_edges):
        (u,v) = uv.copy()
        (k1,k2) = k
        llh = llh0 = 0
        for (kk,ii) in zip([k2] if k1 == 0 else [k1,k2],
                           [u]  if k1 == 0 else [u,v]):
            ii = ii.copy()
            kk = np.where(label_sort == kk)[0][0]
            bds = boundary_distances(k1, k2, theta[ii], eccen[ii], x[ii], y[ii],
                                     uvm_ipsi[kk], lvm_ipsi[kk], max_eccen)
            if bds is None: continue
            llh = llh - bds
        if k1 != 0:
            llh = llh - ((x[u] - x[v])**2 + (y[u] - y[v])**2)
        if llh is llh0: continue
        llhs[k] = llh
        llhtot += torch.sum(llh)
    return (llhtot, pyr.pmap(llhs))
@pimms.calc('likelihood')
def calc_likelihood(dtype, device, mesh, vmag_likelihood, meas_likelihood, boundary_likelihood,
                    model_knob=12, boundary_knob=2):
    '''
    Calculates the likelihood of the model given the measurements.

    Afferent parameters:
     @ model_knob The base-2 log of the constant weight that the vmag_likelihood
       is multiplied by prior to summation with meas_likelihood.
     @ boundary_knob The base-2 log of the constant weight that the
       boundary_likelihood is multiplied by prior to summation with
       meas_likelihood.

    Efferent values:
      @ likelihood The likelihood of the measurements given the model times the
        likelihood of the model.
    '''
    two = astensor(2.0, dtype=dtype, device=device)
    wv = 0 if model_knob is None else two ** model_knob
    wb = 0 if boundary_knob is None else two ** boundary_knob
    return ((meas_likelihood + wv*vmag_likelihood + wb*boundary_likelihood),)
likelihood_plan = pimms.plan(meas_stddev=calc_meas_stddev,
                             meas_likelihood=calc_meas_likelihood,
                             vmag_likelihood=calc_vmag_likelihood,
                             boundary_likelihood=calc_boundary_likelihood,
                             likelihood=calc_likelihood)

# Overall plan
prfclean_plan = pimms.plan(init_params_plan,
                           init_subject_plan,
                           init_ipsi_plan,
                           coords_plan,
                           vmag_plan,
                           model_plan,
                           likelihood_plan)

                           



####################################################################################################
# Models and routines used in visual neuroscience.
# By Noah C. Benson

from .models     import (load_fmm_model, visual_area_names,
                         visual_area_numbers, visual_area_field_signs,
                         RetinotopyModel, RetinotopyMeshModel, RegisteredRetinotopyModel,
                         SchiraModel)
from .retinotopy import (empirical_retinotopy_data, predicted_retinotopy_data, retinotopy_data,
                         extract_retinotopy_argument, retinotopy_comparison, to_logeccen,
                         register_retinotopy, retinotopy_registration, from_logeccen,
                         retinotopy_anchors, retinotopy_model, predict_retinotopy,
                         retinotopy_data, as_retinotopy, retinotopic_field_sign,
                         predict_pRF_radius, fit_pRF_radius, occipital_flatmap,
                         clean_retinotopy_potential, clean_retinotopy, visual_isolines,
                         visual_field_mesh, retinotopic_property_aliases,
                         sectors_to_labels, labels_to_sectors, sector_bounds, refit_sectors)
from .cmag       import (mag_data, is_mag_data, neighborhood_cortical_magnification, face_vmag,
                         face_rtcmag, cmag, areal_cmag, field_of_view, isoline_vmag, disk_vmag)

